import os

import dash
import pandas as pd
from dash.dependencies import Input, Output
from dash import dcc
from dash import html
import dash_pivottable
import plotly.express as px

data = [
    ["Total Bill", "Tip", "Payer Gender", "Payer Smoker", "Day of Week", "Meal", "Party Size"],
    [16.99, 1.01, "Female", "Non-Smoker", "Sunday", "Dinner", 2],
    [10.34, 1.66, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [21.01, 3.5, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [23.68, 3.31, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [24.59, 3.61, "Female", "Non-Smoker", "Sunday", "Dinner", 4],
    [25.29, 4.71, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [8.77, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [26.88, 3.12, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [15.04, 1.96, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [14.78, 3.23, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [10.27, 1.71, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [35.26, 5, "Female", "Non-Smoker", "Sunday", "Dinner", 4],
    [15.42, 1.57, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [18.43, 3, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [14.83, 3.02, "Female", "Non-Smoker", "Sunday", "Dinner", 2],
    [21.58, 3.92, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [10.33, 1.67, "Female", "Non-Smoker", "Sunday", "Dinner", 3],
    [16.29, 3.71, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [16.97, 3.5, "Female", "Non-Smoker", "Sunday", "Dinner", 3],
    [20.65, 3.35, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [17.92, 4.08, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [20.29, 2.75, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [15.77, 2.23, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [39.42, 7.58, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [19.82, 3.18, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [17.81, 2.34, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [13.37, 2, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [12.69, 2, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [21.7, 4.3, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [19.65, 3, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [9.55, 1.45, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [18.35, 2.5, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [15.06, 3, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [20.69, 2.45, "Female", "Non-Smoker", "Saturday", "Dinner", 4],
    [17.78, 3.27, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [24.06, 3.6, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [16.31, 2, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [16.93, 3.07, "Female", "Non-Smoker", "Saturday", "Dinner", 3],
    [18.69, 2.31, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [31.27, 5, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [16.04, 2.24, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [17.46, 2.54, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [13.94, 3.06, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [9.68, 1.32, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [30.4, 5.6, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [18.29, 3, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [22.23, 5, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [32.4, 6, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [28.55, 2.05, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [18.04, 3, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [12.54, 2.5, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [10.29, 2.6, "Female", "Non-Smoker", "Sunday", "Dinner", 2],
    [34.81, 5.2, "Female", "Non-Smoker", "Sunday", "Dinner", 4],
    [9.94, 1.56, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [25.56, 4.34, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [19.49, 3.51, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [38.01, 3, "Male", "Smoker", "Saturday", "Dinner", 4],
    [26.41, 1.5, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [11.24, 1.76, "Male", "Smoker", "Saturday", "Dinner", 2],
    [48.27, 6.73, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [20.29, 3.21, "Male", "Smoker", "Saturday", "Dinner", 2],
    [13.81, 2, "Male", "Smoker", "Saturday", "Dinner", 2],
    [11.02, 1.98, "Male", "Smoker", "Saturday", "Dinner", 2],
    [18.29, 3.76, "Male", "Smoker", "Saturday", "Dinner", 4],
    [17.59, 2.64, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [20.08, 3.15, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [16.45, 2.47, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [3.07, 1, "Female", "Smoker", "Saturday", "Dinner", 1],
    [20.23, 2.01, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [15.01, 2.09, "Male", "Smoker", "Saturday", "Dinner", 2],
    [12.02, 1.97, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [17.07, 3, "Female", "Non-Smoker", "Saturday", "Dinner", 3],
    [26.86, 3.14, "Female", "Smoker", "Saturday", "Dinner", 2],
    [25.28, 5, "Female", "Smoker", "Saturday", "Dinner", 2],
    [14.73, 2.2, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [10.51, 1.25, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [17.92, 3.08, "Male", "Smoker", "Saturday", "Dinner", 2],
    [27.2, 4, "Male", "Non-Smoker", "Thursday", "Lunch", 4],
    [22.76, 3, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [17.29, 2.71, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [19.44, 3, "Male", "Smoker", "Thursday", "Lunch", 2],
    [16.66, 3.4, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [10.07, 1.83, "Female", "Non-Smoker", "Thursday", "Lunch", 1],
    [32.68, 5, "Male", "Smoker", "Thursday", "Lunch", 2],
    [15.98, 2.03, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [34.83, 5.17, "Female", "Non-Smoker", "Thursday", "Lunch", 4],
    [13.03, 2, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [18.28, 4, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [24.71, 5.85, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [21.16, 3, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [28.97, 3, "Male", "Smoker", "Friday", "Dinner", 2],
    [22.49, 3.5, "Male", "Non-Smoker", "Friday", "Dinner", 2],
    [5.75, 1, "Female", "Smoker", "Friday", "Dinner", 2],
    [16.32, 4.3, "Female", "Smoker", "Friday", "Dinner", 2],
    [22.75, 3.25, "Female", "Non-Smoker", "Friday", "Dinner", 2],
    [40.17, 4.73, "Male", "Smoker", "Friday", "Dinner", 4],
    [27.28, 4, "Male", "Smoker", "Friday", "Dinner", 2],
    [12.03, 1.5, "Male", "Smoker", "Friday", "Dinner", 2],
    [21.01, 3, "Male", "Smoker", "Friday", "Dinner", 2],
    [12.46, 1.5, "Male", "Non-Smoker", "Friday", "Dinner", 2],
    [11.35, 2.5, "Female", "Smoker", "Friday", "Dinner", 2],
    [15.38, 3, "Female", "Smoker", "Friday", "Dinner", 2],
    [44.3, 2.5, "Female", "Smoker", "Saturday", "Dinner", 3],
    [22.42, 3.48, "Female", "Smoker", "Saturday", "Dinner", 2],
    [20.92, 4.08, "Female", "Non-Smoker", "Saturday", "Dinner", 2],
    [15.36, 1.64, "Male", "Smoker", "Saturday", "Dinner", 2],
    [20.49, 4.06, "Male", "Smoker", "Saturday", "Dinner", 2],
    [25.21, 4.29, "Male", "Smoker", "Saturday", "Dinner", 2],
    [18.24, 3.76, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [14.31, 4, "Female", "Smoker", "Saturday", "Dinner", 2],
    [14, 3, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [7.25, 1, "Female", "Non-Smoker", "Saturday", "Dinner", 1],
    [38.07, 4, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [23.95, 2.55, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [25.71, 4, "Female", "Non-Smoker", "Sunday", "Dinner", 3],
    [17.31, 3.5, "Female", "Non-Smoker", "Sunday", "Dinner", 2],
    [29.93, 5.07, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [10.65, 1.5, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [12.43, 1.8, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [24.08, 2.92, "Female", "Non-Smoker", "Thursday", "Lunch", 4],
    [11.69, 2.31, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [13.42, 1.68, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [14.26, 2.5, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [15.95, 2, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [12.48, 2.52, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [29.8, 4.2, "Female", "Non-Smoker", "Thursday", "Lunch", 6],
    [8.52, 1.48, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [14.52, 2, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [11.38, 2, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [22.82, 2.18, "Male", "Non-Smoker", "Thursday", "Lunch", 3],
    [19.08, 1.5, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [20.27, 2.83, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [11.17, 1.5, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [12.26, 2, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [18.26, 3.25, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [8.51, 1.25, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [10.33, 2, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [14.15, 2, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [16, 2, "Male", "Smoker", "Thursday", "Lunch", 2],
    [13.16, 2.75, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [17.47, 3.5, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [34.3, 6.7, "Male", "Non-Smoker", "Thursday", "Lunch", 6],
    [41.19, 5, "Male", "Non-Smoker", "Thursday", "Lunch", 5],
    [27.05, 5, "Female", "Non-Smoker", "Thursday", "Lunch", 6],
    [16.43, 2.3, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [8.35, 1.5, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [18.64, 1.36, "Female", "Non-Smoker", "Thursday", "Lunch", 3],
    [11.87, 1.63, "Female", "Non-Smoker", "Thursday", "Lunch", 2],
    [9.78, 1.73, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [7.51, 2, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [14.07, 2.5, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [13.13, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [17.26, 2.74, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [24.55, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [19.77, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [29.85, 5.14, "Female", "Non-Smoker", "Sunday", "Dinner", 5],
    [48.17, 5, "Male", "Non-Smoker", "Sunday", "Dinner", 6],
    [25, 3.75, "Female", "Non-Smoker", "Sunday", "Dinner", 4],
    [13.39, 2.61, "Female", "Non-Smoker", "Sunday", "Dinner", 2],
    [16.49, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [21.5, 3.5, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [12.66, 2.5, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [16.21, 2, "Female", "Non-Smoker", "Sunday", "Dinner", 3],
    [13.81, 2, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [17.51, 3, "Female", "Smoker", "Sunday", "Dinner", 2],
    [24.52, 3.48, "Male", "Non-Smoker", "Sunday", "Dinner", 3],
    [20.76, 2.24, "Male", "Non-Smoker", "Sunday", "Dinner", 2],
    [31.71, 4.5, "Male", "Non-Smoker", "Sunday", "Dinner", 4],
    [10.59, 1.61, "Female", "Smoker", "Saturday", "Dinner", 2],
    [10.63, 2, "Female", "Smoker", "Saturday", "Dinner", 2],
    [50.81, 10, "Male", "Smoker", "Saturday", "Dinner", 3],
    [15.81, 3.16, "Male", "Smoker", "Saturday", "Dinner", 2],
    [7.25, 5.15, "Male", "Smoker", "Sunday", "Dinner", 2],
    [31.85, 3.18, "Male", "Smoker", "Sunday", "Dinner", 2],
    [16.82, 4, "Male", "Smoker", "Sunday", "Dinner", 2],
    [32.9, 3.11, "Male", "Smoker", "Sunday", "Dinner", 2],
    [17.89, 2, "Male", "Smoker", "Sunday", "Dinner", 2],
    [14.48, 2, "Male", "Smoker", "Sunday", "Dinner", 2],
    [9.6, 4, "Female", "Smoker", "Sunday", "Dinner", 2],
    [34.63, 3.55, "Male", "Smoker", "Sunday", "Dinner", 2],
    [34.65, 3.68, "Male", "Smoker", "Sunday", "Dinner", 4],
    [23.33, 5.65, "Male", "Smoker", "Sunday", "Dinner", 2],
    [45.35, 3.5, "Male", "Smoker", "Sunday", "Dinner", 3],
    [23.17, 6.5, "Male", "Smoker", "Sunday", "Dinner", 4],
    [40.55, 3, "Male", "Smoker", "Sunday", "Dinner", 2],
    [20.69, 5, "Male", "Non-Smoker", "Sunday", "Dinner", 5],
    [20.9, 3.5, "Female", "Smoker", "Sunday", "Dinner", 3],
    [30.46, 2, "Male", "Smoker", "Sunday", "Dinner", 5],
    [18.15, 3.5, "Female", "Smoker", "Sunday", "Dinner", 3],
    [23.1, 4, "Male", "Smoker", "Sunday", "Dinner", 3],
    [15.69, 1.5, "Male", "Smoker", "Sunday", "Dinner", 2],
    [19.81, 4.19, "Female", "Smoker", "Thursday", "Lunch", 2],
    [28.44, 2.56, "Male", "Smoker", "Thursday", "Lunch", 2],
    [15.48, 2.02, "Male", "Smoker", "Thursday", "Lunch", 2],
    [16.58, 4, "Male", "Smoker", "Thursday", "Lunch", 2],
    [7.56, 1.44, "Male", "Non-Smoker", "Thursday", "Lunch", 2],
    [10.34, 2, "Male", "Smoker", "Thursday", "Lunch", 2],
    [43.11, 5, "Female", "Smoker", "Thursday", "Lunch", 4],
    [13, 2, "Female", "Smoker", "Thursday", "Lunch", 2],
    [13.51, 2, "Male", "Smoker", "Thursday", "Lunch", 2],
    [18.71, 4, "Male", "Smoker", "Thursday", "Lunch", 3],
    [12.74, 2.01, "Female", "Smoker", "Thursday", "Lunch", 2],
    [13, 2, "Female", "Smoker", "Thursday", "Lunch", 2],
    [16.4, 2.5, "Female", "Smoker", "Thursday", "Lunch", 2],
    [20.53, 4, "Male", "Smoker", "Thursday", "Lunch", 4],
    [16.47, 3.23, "Female", "Smoker", "Thursday", "Lunch", 3],
    [26.59, 3.41, "Male", "Smoker", "Saturday", "Dinner", 3],
    [38.73, 3, "Male", "Smoker", "Saturday", "Dinner", 4],
    [24.27, 2.03, "Male", "Smoker", "Saturday", "Dinner", 2],
    [12.76, 2.23, "Female", "Smoker", "Saturday", "Dinner", 2],
    [30.06, 2, "Male", "Smoker", "Saturday", "Dinner", 3],
    [25.89, 5.16, "Male", "Smoker", "Saturday", "Dinner", 4],
    [48.33, 9, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [13.27, 2.5, "Female", "Smoker", "Saturday", "Dinner", 2],
    [28.17, 6.5, "Female", "Smoker", "Saturday", "Dinner", 3],
    [12.9, 1.1, "Female", "Smoker", "Saturday", "Dinner", 2],
    [28.15, 3, "Male", "Smoker", "Saturday", "Dinner", 5],
    [11.59, 1.5, "Male", "Smoker", "Saturday", "Dinner", 2],
    [7.74, 1.44, "Male", "Smoker", "Saturday", "Dinner", 2],
    [30.14, 3.09, "Female", "Smoker", "Saturday", "Dinner", 4],
    [12.16, 2.2, "Male", "Smoker", "Friday", "Lunch", 2],
    [13.42, 3.48, "Female", "Smoker", "Friday", "Lunch", 2],
    [8.58, 1.92, "Male", "Smoker", "Friday", "Lunch", 1],
    [15.98, 3, "Female", "Non-Smoker", "Friday", "Lunch", 3],
    [13.42, 1.58, "Male", "Smoker", "Friday", "Lunch", 2],
    [16.27, 2.5, "Female", "Smoker", "Friday", "Lunch", 2],
    [10.09, 2, "Female", "Smoker", "Friday", "Lunch", 2],
    [20.45, 3, "Male", "Non-Smoker", "Saturday", "Dinner", 4],
    [13.28, 2.72, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [22.12, 2.88, "Female", "Smoker", "Saturday", "Dinner", 2],
    [24.01, 2, "Male", "Smoker", "Saturday", "Dinner", 4],
    [15.69, 3, "Male", "Smoker", "Saturday", "Dinner", 3],
    [11.61, 3.39, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [10.77, 1.47, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [15.53, 3, "Male", "Smoker", "Saturday", "Dinner", 2],
    [10.07, 1.25, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [12.6, 1, "Male", "Smoker", "Saturday", "Dinner", 2],
    [32.83, 1.17, "Male", "Smoker", "Saturday", "Dinner", 2],
    [35.83, 4.67, "Female", "Non-Smoker", "Saturday", "Dinner", 3],
    [29.03, 5.92, "Male", "Non-Smoker", "Saturday", "Dinner", 3],
    [27.18, 2, "Female", "Smoker", "Saturday", "Dinner", 2],
    [22.67, 2, "Male", "Smoker", "Saturday", "Dinner", 2],
    [17.82, 1.75, "Male", "Non-Smoker", "Saturday", "Dinner", 2],
    [18.78, 3, "Female", "Non-Smoker", "Thursday", "Dinner", 2]
]


def create_dashboard_app(outputs_dir):
    # Prepare dashboard
    df = pd.read_csv(os.path.join(outputs_dir, "indicators.csv"))
    tmp = df.values.tolist()
    tmp.insert(0, df.columns.to_list())
    external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    app.layout = html.Div(children=[
        html.H1(children=f'Enbios (Sentinel-Calliope) dashboard'),
        html.Div(children=[
            html.Div([
                dash_pivottable.PivotTable(
                    id='table',
                    data=tmp,
                    rows=["System", "Processor"],
                    colOrder="key_a_to_z",
                    cols=["Scenario", "Period", "Indicator"],
                    rowOrder="key_a_to_z",
                    rendererName="",
                    aggregatorName="Average",
                    vals=["Value"],
                    valueFilter={'Scope': {'Total': True}}
                ),
            ], className="six columns"),
            html.Div([
                # html.Div(
                #     id='output'
                # )
            ], className="six columns"),
        ], className="row"),
    ])

    @app.callback(Output('output', 'children'),
                  [Input('table', 'cols'),
                   Input('table', 'rows'),
                   Input('table', 'rowOrder'),
                   Input('table', 'colOrder'),
                   Input('table', 'aggregatorName'),
                   Input('table', 'rendererName')])
    def display_props(cols, rows, row_order, col_order, aggregator, renderer):
        return [
            html.P(str(cols), id='columns'),
            html.P(str(rows), id='rows'),
            html.P(str(row_order), id='row_order'),
            html.P(str(col_order), id='col_order'),
            html.P(str(aggregator), id='aggregator'),
            html.P(str(renderer), id='renderer'),
        ]

    return app


import pandas as pd
from typing import List, Tuple


def list_to_dataframe(lst: List) -> pd.DataFrame:
    return pd.DataFrame(data=lst[1:], columns=lst[0])


def generate_import_commands_command(files: List[str]) -> pd.DataFrame:
    lst = [["Workbook", "Worksheets"]]
    for file in files:
        lst.append([file, ""])

    return list_to_dataframe(lst)


"""
PROCESSORS | BLOCKS | ACTIVITIES 
[
 {"name": "",
  "parent": "",
  "group": "",
  "subsystem_type": "local",
  "system": "",
  "function_or_structural": "Structural",
  "accounted": "yes",
  "parent_membership_weight": "",
  "behave_as": "",
  "level": "",
  "description": "",
  "geolocation_ref": "",
  "geolocation_code": "",
  "geolocation_lat_long": "",
  "attributes": {
   "att": "value",
  }
  "interfaces": [
    {"": "",
     "": "",
     "attributes": {
      "att": "value",
     }
     "observations": [
       {
        "": "",
        "attributes": {
         "att": "value"
        }
       }
     ]
  ]
]
"""


def generate_processors_command(blocks_as_processor_templates: List) -> pd.DataFrame:
    lst = [["ProcessorGroup",
            "Processor",
            "ParentProcessor",
            "SubsystemType",
            "System",
            "FunctionalOrStructural",
            "Accounted",
            "ParentProcessorWeight",
            "BehaveAs",
            "Level",
            "Description",
            "GeolocationRef",
            "GeolocationCode",
            "GeolocationLatLong",
            "Attributes"]]
    for b in blocks_as_processor_templates:
        t = [b.get("group", ""),
             b.get("name"),
             b.get("parent"),
             b.get("subsystem_type", ""),
             b.get("system", ""),
             b.get("functional_or_structural", "Structural"),
             b.get("accounted", "Yes"),
             b.get("parent_membership_weight", ""),
             b.get("behave_as", ""),
             b.get("level", ""),
             b.get("description", ""),
             b.get("geolocation_ref", ""),
             b.get("geolocation_code", ""),
             b.get("geolocation_lat_long", ""),
             b.get("attributes", ""),  # Transform into a comma separated list of key=value
             ]
        lst.append(t)

    return list_to_dataframe(lst)


def generate_interfaces_command(blocks_as_processor_templates: List) -> pd.DataFrame:
    lst = [[
        # Interface
        "Processor",
        "InterfaceType",
        "Interface",
        "Sphere",
        "RoegenType",
        "Orientation",
        "OppositeSubsystemType",
        "GeolocationRef",
        "GeolocationCode",
        "Range",
        "RangeUnit",
        "InterfaceAttributes",
        # Quantity
        "Value",
        "Unit",
        "RelativeTo",
        "Uncertainty",
        "Assessment",
        "PedigreeMatrix",
        "Pedigree",
        "Time",
        "Source",
        "NumberAttributes",
        "Comments"
    ]]
    for b in blocks_as_processor_templates:
        processor = f'{b.get("parent")}.{b.get("name")}'
        for i in b.get("interfaces"):
            t = [processor,
                 i.get("interface_type"),
                 i.get("interface"),
                 i.get("sphere", ""),
                 i.get("roegen_type", ""),
                 i.get("orientation"),
                 i.get("opposite_subsystem_type", ""),
                 i.get("geolocation_ref", ""),
                 i.get("geolocation_code", ""),
                 i.get("range", ""),
                 i.get("attributes", ""),  # Transform into a comma separated list of key=value
                 ]
            if len(i["observations"]) > 0:
                for o in i["observations"]:
                    s = [
                        o.get("value"),
                        o.get("unit", ""),
                        o.get("relative_to", ""),
                        o.get("uncertainty", ""),
                        o.get("assessment", ""),
                        o.get("pedigree_matrix", ""),
                        o.get("pedigree", ""),
                        o.get("time"),
                        o.get("source"),
                        o.get("attributes", ""),  # Transform into a comma separated list of key=value
                        o.get("comments", "")
                    ]
                    lst.append(t+s)
            else:
                lst.append(t+[""]*11)

    # Convert to pd.DataFrame
    return list_to_dataframe(lst)



import os

import pytest
import sys
import datetime
import time
import ctypes.util
from typing import Union

# scopes (broader to narrower): session, module, class, function (default)


@pytest.fixture(scope="session")
def chrome_driver():
    """
    Prepare Chrome driver, visible or headless
    :return:
    """
    driver = None  # webdriver.Chrome(options=options)

    yield driver  # Execute tests

    driver.quit()


@pytest.fixture(scope="session")
def mssql_database():
    connection = None  # pyodbc.connect(connect_string)

    yield connection

    connection.close()


@pytest.fixture(scope="session")
def h2_database():
    database = None
    yield database
    # TODO Close connection


@pytest.fixture(scope="function")  # Each test function
def time_set():
    """
    Fixture to KEEP system date after a function assumed to set the LOCAL system date
    This would be used by test setting a specific datetime

    :return:
    """

    t1 = datetime.datetime.now()

    yield t1  # Test execution

    # Reset time (considering the time that the test took to execute)
    dif = datetime.datetime.now() - t1

    # Recover system time
    # set_system_date(t1 + dif)


@pytest.fixture(scope="session")
def base_data():
    # TODO Load base data (using dbUnit or other)
    return





"""
A script to read a list of .spold files and obtain a NIS file with the following commands

* InterfaceTypes
* BareProcessors
* Interfaces

"""
import logging
import operator
import os
import sys
import traceback
from os import listdir
from os.path import isfile, join

from nexinfosys.bin.cli_script import print_issues, PrintColors
from nexinfosys.command_generators import Issue, IType
from nexinfosys.command_generators.parser_ast_evaluators import get_nis_name
from nexinfosys.common.helper import download_file
from io import StringIO
import pandas as pd
from bw2io.importers.ecospold2 import Ecospold2DataExtractor
from enbios.common.helper import list_to_dataframe, generate_json, generate_workbook


def generate_csv(o):
    js = generate_json(o)
    df = pd.read_json(js)
    del df["activity"]
    del df["classifications"]
    del df["properties"]
    del df["loc"]
    del df["production volume"]
    del df["uncertainty type"]
    del df["amount"]  # We are generating a reference of Interface Types. Not specific to a particular Activity
    # TODO There are many duplicate "name"s, for now remove them
    df.drop_duplicates(subset=["name"], inplace=True)
    s = StringIO()
    s.write('# To regenerate this file, execute function "generate_csv" in script "read_lci_data (ENVIRO)"\n')
    df.to_csv(s, index=False)
    return s.getvalue()


class MyEcoSpold2DataExtractor(Ecospold2DataExtractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def extract_properties(cls, exc):
        properties = {}

        for obj in exc.iterchildren():
            if obj.tag.endswith("property"):
                properties[obj.name.text] = {"amount": float(obj.get("amount"))}
                if hasattr(obj, "unitName"):
                    properties[obj.name.text]["unit"] = obj.unitName.text
                if hasattr(obj, "comment"):
                    properties[obj.name.text]["comment"] = obj.comment.text
            elif obj.tag.endswith("compartment"):
                properties["compartment"] = obj.compartment.text
                properties["subcompartment"] = obj.subcompartment.text

        return properties


def read_ecospold_file(f):
    lci = None
    issues = []
    found_name = None
    for name in [f, f"{f}.spold"]:
        try:
            if os.path.exists(name) and os.path.isfile(name):
                found_name = name
                _original_stdout = sys.stdout
                _original_stderr = sys.stderr
                sys.stdout = open(os.devnull, 'w')
                sys.stderr = sys.stdout
                lci = MyEcoSpold2DataExtractor.extract(name, "test", use_mp=False)
                sys.stdout.close()
                sys.stdout = _original_stdout
                sys.stderr = _original_stderr
                lci = lci[0]
                break
        except:
            issues.append(Issue(itype=IType.ERROR, description=f"Exception reading file '{f}': {traceback.format_exc()}"))
            # traceback.print_exc()

    if not found_name:
        issues.append(Issue(itype=IType.ERROR, description=f"File '{f}' (or {f}.spold) not found. Please check '@EcoinventFilename' column in the NIS base file or if the file has not been downloaded into the LCI files folder"))

    return lci, found_name, issues


def lci_to_interfaces_csv(lci_file, csv_file):
    lci_tmp, found_name, issues = read_ecospold_file(lci_file)
    s = generate_csv(lci_tmp["exchanges"])
    with open(csv_file, "wt") as f:
        f.write(s)

"""
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/Electricity, PV, 3kWp, multi-Si.spold")
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/Electricity, PV, production mix.spold")
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/5c5c2277-8df1-4a73-a479-9c14deec9bb1.spold")
s = generate_json(lci_tmp["exchanges"])
with open("/home/rnebot/Downloads/exchanges.json.txt", "wt") as f:
    f.write(s)
s = generate_csv(lci_tmp[0]["exchanges"])
with open("/home/rnebot/Downloads/exchanges.csv", "wt") as f:
    f.write(s)

"""


class SpoldToNIS:
    def __init__(self):
        pass

    @staticmethod
    def _get_spold_files_from_correspondence_file(correspondence_path):  # TODO UNUSED
        df = pd.read_csv(correspondence_path, comment="#")
        # Assume all columns are present: "name", "match_target_type", "match_target", "weight", "match_conditions"
        spolds = []
        ecoinvent_filename_idx = None
        type_idx = None
        name_idx = None
        for idx, col in enumerate(df.columns):
            if col.lower() == "match_target":
                ecoinvent_filename_idx = idx
            elif col.lower() == "match_target_type":
                type_idx = idx
            elif col.lower() == "name":
                name_idx = idx
        if ecoinvent_filename_idx is not None and type_idx is not None and name_idx is not None:
            for idx, r in df.iterrows():
                ecoinvent_filename = r[df.columns[ecoinvent_filename_idx]]
                o_type = r[df.columns[type_idx]]
                name = r[df.columns[name_idx]]
                if ecoinvent_filename != "" and o_type != "" and name != "":
                    if o_type.lower() in ["lca", "lci"]:
                        spolds.append(dict(name=name, ecoinvent_filename=ecoinvent_filename))
        return spolds

    @staticmethod
    def _get_spold_files_from_nis_file(nis_url):
        """
        Elaborate a list of Spold files (to be later processed) coming from a NIS formatted XLSX file (nis_url).

        The file is examined, looking for columns (row 1) named "processor", "@ecoinventfilename" and "@ecoinventname"
        in "BareProcessors" sheets to construct each of the entries (each entry is a "dict") of the returned list.

        :param nis_url: Location of NIS formatted XLSX
        :return: List of dict's, where each dict contains the information of a Spold file
        """
        bytes_io = download_file(nis_url)
        xl = pd.ExcelFile(bytes_io.getvalue(), engine='openpyxl')
        # xl = pd.ExcelFile(xlrd.open_workbook(file_contents=bytes_io.getvalue()), engine="xlrd")
        spolds = []
        for sheet_name in xl.sheet_names:
            if not sheet_name.lower().startswith("bareprocessors"):
                continue
            df = xl.parse(sheet_name, header=0)
            ecoinvent_filename_idx = None
            ecoinvent_carrier_name_idx = None
            name_idx = None
            for idx, col in enumerate(df.columns):
                if col.lower() == "@ecoinventfilename":
                    ecoinvent_filename_idx = idx
                elif col.lower() == "@ecoinventcarriername":
                    ecoinvent_carrier_name_idx = idx
                elif col.lower() == "processor":
                    name_idx = idx
            if ecoinvent_filename_idx is not None and name_idx is not None:
                for idx, r in df.iterrows():
                    ecoinvent_filename = r[df.columns[ecoinvent_filename_idx]]
                    if ecoinvent_carrier_name_idx is not None:
                        ecoinvent_carrier_name = r[df.columns[ecoinvent_carrier_name_idx]]
                    else:
                        ecoinvent_carrier_name = None
                    name = r[df.columns[name_idx]]
                    if ecoinvent_filename != "" and not isinstance(ecoinvent_filename, float) and name != "":
                        spolds.append(dict(name=name,
                                           ecoinvent_filename=ecoinvent_filename,
                                           ecoinvent_carrier_name=ecoinvent_carrier_name))
        return spolds

    def spold2nis(self, default_output_interface: str, lci_base: str, correspondence, nis_base_url: str, output_file: str):
        """
        A method to transform Spold files into a NIS Workbook with InterfacesTypes, BareProcessors and Interfaces

        :param default_output_interface: Name of the interface to use when no clear output interface is found in the .spold file
        :param lci_base: Local path where .spold files are located
        :param correspondence: Location of the correspondence file (open "sim_correspondence_example.csv" file)
        :param nis_base_url: Location of the NIS Base file, which can also have processors with LCI location
        :param output_file: Path of the resulting Workbook
        :return: None
        """

        # Read exchanges into a pd.DataFrame (to produce InterfaceTypes and Interfaces)
        def _lci_to_dataframe(lci_in):
            try:
                j = generate_json(lci_in["exchanges"])
                df_tmp = pd.read_json(StringIO(j))
            except:
                traceback.print_exc()
                logging.debug(f"Problem converting '{file_name}' 'exchanges' to JSON. Skipped.")
                return None, None

            for c in ["activity", "classifications", "properties", "loc", "production volume", "uncertainty type"]:
                del df_tmp[c]
            dfi_tmp = df_tmp.copy()
            del dfi_tmp["amount"]
            # There are duplicate "name"s, for now remove them
            dfi_tmp.drop_duplicates(subset=["name"], inplace=True)
            return df_tmp, dfi_tmp

        def _find_main_output(aggregate_exchanges):
            """ Find main output """
            _main_output = None
            _main_output_is_output = None
            for _idx, _r in aggregate_exchanges.iterrows():
                if _r["amount"] == 1.0:
                    _main_output = get_nis_name(_r["name"])
                    _main_output_is_output = True
                    break
            if _main_output is None:
                # Special activities can be scaled according to a -1 exchange
                # (Nick clarified this unusual but possible case)
                for _idx, _r in aggregate_exchanges.iterrows():
                    if _r["amount"] == -1.0:
                        _main_output = get_nis_name(_r["name"])
                        _main_output_is_output = False
                        break
            if _main_output is None:
                _main_output = default_output_interface
                _main_output_is_output = True

            return _main_output, _main_output_is_output

        def _generate_commands(iface_types, procs, ifaces):
            """ Generate the three commands, InterfaceTypes, BareProcessors, Interfaces """
            _cmds = []

            # InterfaceTypes
            header = ["InterfaceTypeHierarchy", "InterfaceType", "Sphere", "RoegenType", "ParentInterfaceType",
                      "Formula",
                      "Description", "Unit", "OppositeSubsystemType", "Attributes", "@EcoinventName"]
            lst = []
            if default_output_interface != "":
                lst.append(["LCI", default_output_interface, "Technosphere", "Flow", "", "",
                            "Default-generic output for LCI activities not stating explicitly its output interface",
                            "EJ", "", "", ""])
            for interface_type, props in iface_types.items():
                opposite = "Environment" if props["sphere"].lower() == "biosphere" else ""
                lst.append(["LCI", interface_type, props["sphere"], "Flow", "", "",
                            props["comment"], props["unit"], opposite, "", props["lci_name"]])
            lst = sorted(lst, key=operator.itemgetter(1))
            lst.insert(0, header)
            _cmds.append(("InterfaceTypes", list_to_dataframe(lst)))

            # BareProcessors
            lst = [
                ["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
                 "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong",
                 "Attributes", "@EcoinventName", "@EcoinventFilename", "@EcoinventCarrierName", "@region"]]
            for processor, props in procs.items():
                spold_file = props["lca_file"]
                lst.append(
                    ["EcoinventReferenceStructurals", processor, "", "", "", "Structural", "No", "", "", "", "", "",
                     "", props["name"], spold_file, props["carrier_name"], ""])
            _cmds.append(("BareProcessors", list_to_dataframe(lst)))

            # Interfaces
            lst = [
                ["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation",
                 "OppositeSubsystemType",
                 "GeolocationRef", "GeolocationCode", "I@compartment", "I@subcompartment",
                 "Value", "Unit", "RelativeTo",
                 "Uncertainty",
                 "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
            for p_i, props in ifaces.items():
                v = props["value"]
                orientation = "Output" if props["is_output"] else "Input"
                output_name = props["relative_to"]
                iface = p_i[2] if len(p_i) > 2 else ""
                comp = props["compartment"] if "compartment" in props else ""
                subcomp = props["subcompartment"] if "subcompartment" in props else ""
                lst.append([p_i[0], p_i[1], iface, "", "", orientation,
                            "",
                            "", "", comp, subcomp,
                            v, "", output_name,
                            "",
                            "", "", "", "Year", "Ecoinvent", "", ""])
            _cmds.append(("Interfaces", list_to_dataframe(lst)))
            return _cmds

        # FUNCTION STARTS HERE -----------------
        issues = []

        # Read the list of Spold files to process
        spolds = []
        if correspondence:
            _ = self._get_spold_files_from_correspondence_file(correspondence)
            spolds.extend(_)
        if nis_base_url:
            _ = self._get_spold_files_from_nis_file(nis_base_url)
            spolds.extend(_)

        interface_types = {}
        processors = {}
        interfaces = {}

        # Process each .spold file
        files_in_lci_base = set([join(lci_base, f) for f in listdir(lci_base) if isfile(join(lci_base, f))])
        already_processed = set()
        for spold in spolds:
            file_name = os.path.join(f"{lci_base}", f"{spold['ecoinvent_filename']}")
            if file_name in already_processed:
                continue
            else:
                already_processed.add(file_name)

            # Read Spold file
            lci, found_name, _ = read_ecospold_file(file_name)
            issues.extend(_)
            if lci is None:
                continue
            if found_name:
                issues.append(Issue(itype=IType.INFO, description=f"LCI file '{os.path.basename(found_name)}' read correctly"))
                files_in_lci_base.remove(found_name)

            # Bring compartment and subcompartment to exchange properties level
            for exc in lci["exchanges"]:
                exc["compartment"] = exc["properties"].get("compartment")
                exc["subcompartment"] = exc["properties"].get("subcompartment")

            # Transform Spold file into usable data structures (pd.DataFrame's)
            df, dfi = _lci_to_dataframe(lci)
            if df is None:
                continue

            # Obtain processor name
            p_name = lci["name"]

            # PROCESSORS
            nis_name = get_nis_name(p_name)
            if nis_name in processors:
                cont = 1
                while f"{nis_name}_{cont}" in processors:
                    cont += 1
                nis_name = f"{nis_name}_{cont}"
            processors[nis_name] = dict(name=p_name, lca_code=lci["activity"],
                                        lca_type=lci["activity type"], lca_file=lci["filename"],
                                        carrier_name=spold.get("ecoinvent_carrier_name", ""),
                                        description=lci["comment"])

            # INTERFACE TYPES <- dfi
            for idx, r in dfi.iterrows():
                sphere = "Technosphere" if r["type"].lower() != "biosphere" else "Biosphere"
                interface_types[get_nis_name(r["name"])] = dict(comment=r.get("comment", ""),
                                                                flow=r["flow"],
                                                                sphere=sphere,
                                                                unit=r["unit"],
                                                                lci_name=r["name"])

            # INTERFACES <- df

            # Find main output
            main_output, main_output_is_output = _find_main_output(df)

            # Add output interface first
            i_name = get_nis_name(main_output)
            interfaces[(nis_name, i_name)] = dict(value=1, relative_to="", is_output=main_output_is_output)
            # Add rest of interfaces
            for idx, r in df.iterrows():
                compartment = r["compartment"]
                subcompartment = r["subcompartment"]
                i_name = get_nis_name(r["name"])
                it_name = i_name
                if compartment:
                    i_name += f"_{get_nis_name(compartment)}"
                if subcompartment:
                    i_name += f"_{get_nis_name(subcompartment)}"
                if (nis_name, i_name) not in interfaces:
                    relative_to = main_output if i_name != main_output else ""
                    # value = r["amount"] if relative_to != "" else ""
                    interfaces[(nis_name, it_name, i_name)] = dict(value=r["amount"],
                                                                   relative_to=relative_to,
                                                                   compartment=compartment,
                                                                   subcompartment=subcompartment,
                                                                   is_output=i_name == main_output)
        for f in files_in_lci_base:
            issues.append(Issue(itype=IType.WARNING, description=f"File '{os.path.basename(f)}' not used"))

        # Generate InterfaceTypes, BareProcessors and Interfaces commands
        cmds = _generate_commands(interface_types, processors, interfaces)

        # Generate an in-memory XLSX file
        s = generate_workbook(cmds)

        # Write to file
        if s:
            with open(output_file, "wb") as f:
                f.write(s)

        # Print collected issues
        print_issues("Spold2NIS", "", issues, f"Please check the issues and output file '{output_file}'")

        if s:
            print(f"\nNIS file:\n"
                  f"{PrintColors.BOLD}{output_file}{PrintColors.END}\n"
                  f"containing information from LCI files generated successfully.\n"
                  f"Please, refer to this file from the NIS Base file, using an ImportCommands command (worksheet).\n"
                  f"If NIS Base file is a Google Sheets document:\n"
                  f" * Upload the generated file to Google Drive,\n"
                  f" * Go to the uploaded file in Google Drive, open right click menu, 'Get Link' (select 'anybody with the link') and\n"
                  f" * Copy/paste the link into the appropriate 'ImportCommands' cell (under 'Workbook' column) of NIS base file.")



from io import StringIO
from typing import List

import openpyxl
import pandas as pd
from nexinfosys.command_generators.parser_ast_evaluators import get_nis_name


def convert_lcia_implementation_to_nis(lcia_implementation_file: str, lcia_file: str,
                                       method_like, method_is: List[str], include_obsolete, use_nis_name_syntax):
    def get_horizon(ind):
        return ""

    def adapt_name(name):
        return get_nis_name(name) if use_nis_name_syntax else name

    workbook = openpyxl.load_workbook(lcia_implementation_file, data_only=True)
    cfs = workbook["CFs"]
    units = workbook["units"] if "units" in workbook.sheetnames else workbook["Indicators"]
    methods = set()
    categories = set()
    indicators = {}  # And its unit
    horizons = set()
    compartments = {}  # Include subcompartments
    if method_like != "":
        method_like = method_like.lower()
    if method_is:
        method_is = [m.lower() for m in method_is]

    for r in range(2, units.max_row):
        method = units.cell(row=r, column=1).value.strip()
        if not include_obsolete:
            if "superseded" in method.lower() or "obsolete" in method.lower():
                continue
        if method_like != "":
            if method_like not in method.lower():
                continue
        if method_is is not None:
            if method.lower() not in method_is:
                continue

        category = units.cell(row=r, column=2).value.strip()
        indicator = units.cell(row=r, column=3).value.strip()
        unit = units.cell(row=r, column=4).value.strip()

        methods.add(method)
        categories.add(category)
        indicators[indicator] = unit

    print("METHODS")
    print(list(methods))
    print("INDICATORS")
    print(list(indicators.keys()))
    print("CATEGORIES")
    print(list(categories))
    _ = []  # Output
    for r in range(2, cfs.max_row):
        method = cfs.cell(row=r, column=1).value.strip()
        if method not in methods:
            continue
        category = cfs.cell(row=r, column=2).value.strip()
        indicator = cfs.cell(row=r, column=3).value.strip()
        # Guess horizon
        h = get_horizon(indicator)
        name = cfs.cell(row=r, column=4).value.strip()
        compartment = cfs.cell(row=r, column=5).value.strip()
        subcompartment = cfs.cell(row=r, column=6).value.strip()
        v = cfs.cell(row=r, column=7).value
        ind_unit = indicators[indicator]
        if not include_obsolete:
            if "superseded" in name.lower() or "obsolete" in name.lower():
                continue

        _.append((method,
                  category,
                  indicator,
                  ind_unit,
                  adapt_name(name),  # InterfaceType name
                  "",  # This assumes unit from the InterfaceType definition, which may not be correct
                  h,  # Horizon, which may be empty ("")
                  v,
                  compartment,
                  subcompartment)
                 )

    df = pd.DataFrame(_, columns=["LCIAMethod", "LCIACategory", "LCIAIndicator", "LCIAIndicatorUnit",
                                  "Interface", "InterfaceUnit",
                                  "LCIAHorizon", "LCIACoefficient",
                                  "Compartment", "Subcompartment"])

    s = StringIO()
    s.write(f"# To regenerate this file, execute 'enbios lcia_implementation_to_nis {lcia_implementation_file} {lcia_file} {method_like if method_like else ''} {'--include-obsolete' if include_obsolete else ''}'\n")
    df.to_csv(s, index=False)
    with open(lcia_file, "wt") as f:
        f.write(s.getvalue())


if __name__ == '__main__':
    convert_lcia_implementation_to_nis("/home/rnebot/GoogleDrive/AA_SENTINEL/LCIA_implementation_3.7.1.xlsx", "/home/rnebot/Downloads/test.csv")



import pandas as pd
from nexinfosys.embedded_nis import NIS
from typing import Dict
from enbios.input import Simulation
from enbios.input.lci import LCIIndex
from enbios.processing.nis_file_generator import generate_import_commands_command, generate_processors_command, \
    generate_interfaces_command


def block_meets_conditions(block, condition):
    """
    Parse expression with
    :param block:
    :param condition:
    :return:
    """
    # TODO arith_boolean_expression_with_less_tokens (DESPUÉS DE NEXINFOSYS V0.47)
    from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression
    from nexinfosys.model_services import State
    from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator
    try:
        ast = string_to_ast(arith_boolean_expression, condition)
    except:
        print(f"A syntax error here")
    s = State()
    for k, v in block.items():
        s.set(k, v)
    issues = []
    res, unres = ast_evaluator(ast, s, None, issues)
    if isinstance(res, bool):
        return res
    else:
        return unres  # It is an error probably


class Matcher:
    """
    * Match Activities (LCA) and Technologies (simulation)
    * Match Technologies (simulation) into parent Processors (MuSIASEM)
    """
    def __init__(self, block_types_path, correspondence_path):
        self._block_types_path = block_types_path
        self._correspondence_path = correspondence_path
        self._typed_block = {}
        self._sim_lca = {}  # Store a list of target LCA activities for each simulation type
        self._sim_musiasem = {}  # Store a list of target MuSIASEM processors for each simulation type
        self._read()

    def _read(self):
        """ Read the correspondence file(s). Format:
technology,match_target_type,match_target,weight,match_conditions
wind_onshore_competing,lca,2fdb6da1-8281-453d-8a7c-0184dc3586c4_66c93e71-f32b-4591-901c-55395db5c132.spold,,
wind_onshore_competing,musiasem,Energy_system.Electricity_supply.Electricity_generation.Electricity_renewables.Electricity_wind.Electricity_wind_onshore,,
...
        """
        # ### Read block types (if defined) ###
        self._typed_block = {}
        df = pd.read_csv(self._block_types_path)
        # Assume all columns are present: "name", "type"
        for idx, r in df.iterrows():
            if self._typed_block.get(r["name"]):
                print(f"Entry for {r['name']} repeated. Overwriting")
            self._typed_block[r["name"]] = r["type"]

        # ### Read correspondence ###
        self._sim_lca = {}
        self._sim_musiasem = {}
        df = pd.read_csv(self._correspondence_path)
        # Assume all columns are present: "name", "match_target_type", "match_target", "weight", "match_conditions"
        for idx, r in df.iterrows():
            if r["match_target_type"].lower() == "lca":
                d = self._sim_lca
            else:
                d = self._sim_musiasem
            lst = d.get(r["name"], [])
            if len(lst) == 0:
                d[r["name"]] = lst
            lst.append(dict(target=r["match_target"], weight=r["weight"], conditions=r["match_conditions"]))

    def find_related_lci_activities(self, block: Dict[str, str], lci_index: LCIIndex):
        block_name = block["name"]
        lst = self._sim_lca.get(block_name)
        if not lst:
            block_type = self._typed_block.get(block)
            if block_type:
                return self.find_related_lci_activities(block, lci_index)
        else:
            lst2 = []
            for b in lst:
                if block_meets_conditions(block, b["match_conditions"]):
                    lst2.append(b)
            return [dict(activity=lci_index.read_activity(b["target"]), weight=b["weight"]) for b in lst2]

    def find_candidate_musiasem_parent_processors(self, nis: NIS, block):
        # TODO Find parent processors names in the correspondence, locate them in NIS and return the full hierarchy name
        #      of each
        pass

    def equivalent_interface_type_for_lci_flow(self, lca_flow: str, nis: NIS):
        pass

    def equivalent_interface_type_for_simulation_flow(self, simulation_flow: str, nis: NIS):
        pass



"""
Read both Calliope model and Calliope outputs

* Model into Graph
* Outputs into same Graph
  - dispatch decisions. Useful to evaluate consumption of resource. But compute only accumulated consumption instead of the Time Series
  - capacities. Useful to scale processors/activities


MODEL -> RUN -> ASSESS (read model and results; match model to data; elaborate NIS and BW input; run NIS and BW; get results)


"""

import networkx as nx
import calliope
from typing import List

from enbios.input import Simulation


def read_calliope_model(model_file: str, outputs: str) -> List[nx.DiGraph]:
    """
    Read a Calliope model and return a graph of technologies/sites per scenario
    Also return Calliope model outputs and integrate them into the Graph

    :param base_path:
    :return:
    """
    # Read locations, techs, link-techs, and scenarios (different models)
    # "import" enumerates other files to import, recursively
    model = calliope.Model(model_file)


def map_calliope_to_available(calliope: nx.DiGraph, technologies):
    """
    Although the idea is to obtain real technologies instead of generic ones, based on which LCA assessment can be performed,
    maybe there is more than one technology matching the one modeled.
    The matching criteria could be: location, time, technology type, range (large, medium, small instance), HA, surface, price, ...


    :param calliope:
    :param technologies:
    :return:
    """


class CalliopeSimulation(Simulation):
    """
    Abandoned by Calliope during Sentinel project, in favor of Friendly-Data format, see "sentinel.py"
    """
    def __init__(self, simulation_files_path):
        self._model = None
        self._simulation_files_path = simulation_files_path
        self.read()

    def read(self, filter_model: str):
        self._model = calliope.Model(self._simulation_files_path)
        ds = self._model.inputs
        # Carriers -> InterfaceTypes
        carriers = dict()
        for i in ds.carriers:
            s = str(i.values)
            carriers[s] = dict()

        # Locations -> GeographicalReferences
        locations = dict()
        for i in ds.locs:
            s = str(i.values)
            locations[s] = [ds.loc_coordinates.sel(coordinates=gc, locs=s) for gc in ["lat", "lon"]]
        # Technologies -> Processors (Archetype)
        techs = dict()  # TODO: is Transmission, is Conversion, type
        for i in ds.techs:
            s = str(i.values)
            ttype = None
            techs[s] = dict(tech_type=ttype)

        # Technologies at locations -> Processors (Instances) (need to clone Archetypes? seems no)
        # - Type
        # - Efficiency
        # - Land Use
        # - Power Capacity
        techs_locs = dict()
        for i in ds.loc_techs:
            location, tech = str(i.values).split("::")
            techs_locs[f"{location}__{tech}"] = dict(loc=location, tech=tech)
        print("hola")

        # Transmission are technologies -> Processors, ¿allocated to two locations? ¿proportion 50/50?
        links = dict()
        for i in ds.loc_techs_transmission:
            print(i)

        # TODO
        #  * Read locations (and geographical coordinates)
        #  * Read techs (and general features)
        #  * Read pairs location-tech (and particular features)
        #  * Reading features may imply (or not) going through input CSV files

        # TODO f"{tech}_at_{loc}" -> processor: where? lat, long?

        # TODO Can Calliope read the output of the optimization? or is it necessary to read separately...

    def blocks(self):
        """ List of technologies """

    def block_information(self, block_name):
        """ Return information for block """
        pass




g_default_subtech = "_"  # Constant for default subtech


class SimStructuralProcessorAttributes:
    """
    A class to store Processor attributes while a Sentinel datapackage is read
    and values are stored in a PartialRetrievalDictionary
    Attributes can be stored directly
    """
    def __init__(self, technology=None, region=None, carrier=None, scenario=None, time_=None,
                 subtechnology=None, subscenario=None):
        self.attrs = {}
        if technology:
            self.attrs["technology"] = technology
        if subtechnology:
            self.attrs["subtechnology"] = subtechnology
        if region:
            self.attrs["region"] = region
        if carrier:
            self.attrs["carrier"] = carrier
        if scenario:
            self.attrs["scenario"] = scenario
        if subscenario:
            self.attrs["subscenario"] = subscenario
        if time_:
            self.attrs["time"] = time_

    @staticmethod
    def partial_key(technology=None, region=None, carrier=None, scenario=None, time=None,
                    subtechnology=None, subscenario=None):
        d = {}
        if technology:
            d["_t"] = technology
        if subtechnology:
            d["_st"] = subtechnology
        if region:
            d["_g"] = region
        if carrier:
            d["_c"] = carrier
        if scenario:
            d["_s"] = scenario
        if subscenario:
            d["_ss"] = subscenario
        if time:
            d["_d"] = time
        return d

    def key(self):
        return self.partial_key(self.attrs.get("technology"),
                                self.attrs.get("region"),
                                self.attrs.get("carrier"),
                                self.attrs.get("scenario"),
                                self.attrs.get("time"),
                                self.attrs.get("subtechnology"),
                                self.attrs.get("subscenario"))




# /usr/bin/env python
import os
from typing import List

import fire
import webbrowser

from nexinfosys.bin.cli_script import set_log_level_from_cli_param, prepare_base_state, print_issues
from nexinfosys.command_generators.parser_ast_evaluators import get_nis_name
from nexinfosys.common.helper import PartialRetrievalDictionary

from enbios.common.helper import list_to_dataframe, generate_workbook
from enbios.input.data_preparation.lci_to_nis import SpoldToNIS
from enbios.input.data_preparation.lcia_implementation_to_nis import convert_lcia_implementation_to_nis
from enbios.input.data_preparation.recipe_to_nis import convert_recipe_to_nis
from enbios.input.data_preparation.sentinel_to_nis_prep import sentinel_to_prep_file
from enbios.processing import read_parse_configuration
from enbios.processing.main import Enviro
from enbios.visualize import create_dashboard_app

"""
DEVELOPMENT CLI EXECUTION:
 
python -m enbios.bin.script
  <it will show the automatically generated help>

CLI EXECUTION ("enbios" because it is defined with that name in setup.py):
 
enbios (same as previous)
  <it will show the -automatically generated- help>
  
"""


def list_lcia_indicators(nis_file, output_file, compartment, subcompartment, horizon):

    def query_lcia_methods(state):
        """
        Return commands informing of the available LCIA indicators

        :return:
        """
        lcia_methods_dict = state.get("_lcia_methods")
        if lcia_methods_dict:
            lcia_methods = PartialRetrievalDictionary()
            from random import random
            for k, v in lcia_methods_dict.items():
                # [0] m=Method,
                # [1] t=caTegory,
                # [2] d=inDicator,
                # [3] h=Horizon,
                # [4] i=Interface,
                # [5] c=Compartment,
                # [6] s=Subcompartment,
                _ = dict(m=k[0], t=k[1], d=k[2], h=k[3], i=k[4], c=k[5], s=k[6])
                # NOTE: a random() is generated just to grant that the tuple is unique
                lcia_methods.put(_, (v[0], v[1], v[2], random()))

            methods = lcia_methods.get(dict(m=None, d=None, h=None, c=None, s=None, t=None), just_key=True)

            # Convert to set of tuples, keeping only the desired elements of dictionaries in each tuple, then sort
            keys = ["m", "t", "d"]
            if compartment:
                keys.append("c")
            if subcompartment:
                keys.append("s")
            if horizon:
                keys.append("h")

            subc_str = "" if compartment else "subcompartment="

            different = sorted(set([tuple(map(lambda ke: meth[ke], keys)) for meth in methods]))
            # different = sorted(set([(key["m"], key["t"], key["d"], key["c"], key["s"], key["h"]) for key in keys]),
            #                    key=lambda x: (x[0], x[1], x[2], x[3], x[4], x[5]))
            # Prepare pseudo "ScalarIndicators" command
            lst = [["IndicatorsGroup", "Indicator", "Local", "Processors", "Formula", "Unit", "AccountNA", "UnitLabel", "Benchmarks",
                    "Description", "Reference"]]
            for t in different:
                # Obtain the unit of the indicator
                _ = {}
                for i, d in enumerate(keys):
                    if t[i] is not None:
                        _[d] = t[i]
                o = lcia_methods.get(_)
                if len(o) > 0:
                    unit = o[0][0]
                else:
                    unit = "dimensionless"

                indicator_name = f"{t[0]}_{t[1]}_{t[2]}"
                formula = f'LCIAMethod("{t[0]}", "{t[1]}", "{t[2]}"'
                description = f"Generated by enbios 'lcia_list_indicators', using a base NIS file " \
                              f"where the LCIA methods are defined. " \
                              f"Method {t[0]}, category {t[1]}, indicator {t[2]}"

                if compartment:
                    formula += f', "{t[3]}"'
                    indicator_name += f"_{t[3]}"
                    description += f", compartment {t[3]}"
                    offset = 0
                    f_part = ""
                else:
                    offset = 1
                    f_part = "_"
                    description += f", no compartment"

                if subcompartment:
                    formula += f', {subc_str}"{t[4-offset]}"'
                    indicator_name += f"{f_part}_{t[4-offset]}"
                    description += f", {subc_str}{t[4-offset]}"
                else:
                    offset += 1
                    f_part += "_"
                    description += f", no subcompartment"

                if horizon:
                    formula += f', horizon="{t[5-offset]}"'
                    indicator_name += f"{f_part}_{t[5-offset]}"
                    description += f", horizon {t[5-offset]}"
                else:
                    description += f", no horizon"

                formula += ")"
                description += "."

                # Produce a row
                lst.append(["LCIA",
                            get_nis_name(indicator_name),  # Convert to a NIS name (requested by Miquel, 29/9/2022)
                            "Yes",
                            "",  # Processors ("": "all processors")
                            formula,
                            unit,  # LCIA indicator unit
                            # TODO "Yes" (below) not implemented for LCIAMethod
                            "",  # AccountNA? Yes or "" (for No). "NA" stands for "Not Available"
                            unit,  # UnitLabel
                            "",  # Benchmarks
                            description,
                            ""
                            ])
            return [("ScalarIndicators auto LCIA", list_to_dataframe(lst))]
        else:
            return []

    dir, basename = os.path.split(output_file)
    state, serial_state, issues = prepare_base_state(nis_file, False, dir)
    print_issues("NIS base preparation", nis_file, issues, f"Please check the issues and output file '{output_file}'")
    if state:
        cmds = query_lcia_methods(state)
        s = generate_workbook(cmds, True)
        with open(output_file, "wb") as f:
            f.write(s)


class Enbios:
    @staticmethod
    def recipe_to_csv(recipe_file: str, lcia_file: str, log: str = None):  # TODO DEPRECATED, use "lcia_implementation_to_csv"
        """
        Convert an XLSX file with ReCiPe2016 to a CSV file ready to be declared in a DatasetDef which can be imported
        in a "LCIAMethod" command

        Example:
           enbios recipe_to_csv /home/rnebot/GoogleDrive/AA_SENTINEL/workspace_20210331/ReCiPe2016_CFs_v1.1_20180117.xlsx /home/rnebot/GoogleDrive/AA_SENTINEL/case_studies/library/recipe2016_2.csv


        :param recipe_file: The full path of the input ReCiPe2016 XLSX file
        :param lcia_file: The full path of the output CSV file
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        convert_recipe_to_nis(recipe_file, lcia_file)

    @staticmethod
    def lcia_implementation_to_csv(lcia_implementation_file: str,
                                   lcia_file: str,
                                   method_like: str = "",
                                   method_is: List[str] = None,
                                   include_obsolete: bool = False,
                                   use_nis_name_syntax: bool = True,
                                   log: str = None):
        """
        Convert an XLSX file with LCIA implementation (from Ecoinvent) to a CSV file ready to be declared in a
        DatasetDef which can later be imported in a "LCIAMethod" command

        Example:
           enbios lcia_implementation_to_csv /home/rnebot/GoogleDrive/AA_SENTINEL/LCIA_implementation_3.7.1.xlsx /home/rnebot/GoogleDrive/AA_SENTINEL/case_studies/library/lcia_implementation_nis.csv

           Just one method, using "method-is" ("method-like" is for methods containing the string):
           enbios lcia_implementation_to_csv --method-is "ReCiPe Midpoint (H) V1.13" /home/rnebot/GoogleDrive/AA_SENTINEL/dev_lcia_methods/LCIA_implementation_3.7.1.xlsx "/home/rnebot/GoogleDrive/AA_SENTINEL/dev_nis_case_studies/library (Base)/reduced_lcia_methods_recipe_h_v113.csv"

        :param lcia_implementation_file: The full path of the INPUT LCIA implementation XLSX file
        :param lcia_file: The full path of the OUTPUT CSV file
        :param method_like: To filter rows, let pass only those where method contains "method_like"
        :param method_is: To filter rows, let pass only those where method is in "method_is"
        :param include_obsolete: To not filter rows, where method is considered obsolete
        :param use_nis_name_syntax: Transform names which are used as NIS identifiers to the NIS identifier syntax
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        if isinstance(method_is, str):
            method_is = [method_is]
        convert_lcia_implementation_to_nis(lcia_implementation_file, lcia_file,
                                           method_like, method_is,
                                           include_obsolete, use_nis_name_syntax)

    # def lci_to_interface_types(self, lci_file_path: str, csv_file: str):
    #     """
    #     Extract the interface types from an LCI file
    #
    #     :param lci_file_path: Path of a single .spold file
    #     :param csv_file: Output CSV with a listing of distinct interface types (potentially importable in InterfaceTypes)
    #     :return:
    #     """
    #     lci_to_interfaces_csv(lci_file_path, csv_file)

    @staticmethod
    def lci_to_nis(nis_base_url: str,
                   spold_files_folder: str,
                   nis_structurals_output: str, log: str = None):
        """
        Scan "NIS base file" for references to Spold files (that should be available in "Spold files folder"), and
        elaborate another NIS file at "NIS structurals output", with equivalent MuSIASEM structural Processors.
        "enbios enviro" needs the information to be incorporated into the NIS base file, so the NIS Base file should be
        modified to import the elaborated NIS file, using an ImportCommands command.

        :param spold_files_folder: Local folder where .spold files are stored
        :param nis_base_url: URL of a NIS file that will be used as Base for the assembly of the model
        :param nis_structurals_output: File name of the output NIS that will contain the structural processors
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        # The resulting NIS file has these commands:
        #    - InterfaceTypes
        #    - BareProcessors
        #    - Interfaces

        set_log_level_from_cli_param(log)
        s2n = SpoldToNIS()


        # TODO CORRESPONDENCE_PATH not USED (NIS base file contains all the information)
        #  The correspondence file should be a CSV file with the following header:
        #     name,match_target_type,match_target,weight,match_conditions
        #  Example:
        #     enbios lci_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/enviro_tmp/ /home/rnebot/GoogleDrive/AA_SENTINEL/enviro/sim_correspondence_example.csv https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit?usp=sharing /home/rnebot/Downloads/out.xlsx
        #     enbios lci_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/enviro_tmp/ "" https://docs.google.com/spreadsheets/d/1nYzphq1XYrezquW3yj7UiJ8sNJ8RzJ8gqErg3oVmw2Q/edit?usp=sharing /home/rnebot/Downloads/lci_to_nis_output.xlsx
        #
        #  Example lines in this file:
        #     wind_onshore_competing,lca,2fdb6da1-8281-453d-8a7c-0184dc3586c4_66c93e71-f32b-4591-901c-55395db5c132.spold,,
        #     wind_onshore_competing,musiasem,Energy_system.Electricity_supply.Electricity_generation.Electricity_renewables.Electricity_wind.Electricity_wind_onshore,,

        correspondence_path = ""
        s2n.spold2nis("generic_energy_production",
                      spold_files_folder,
                      correspondence_path,
                      nis_base_url,
                      nis_structurals_output)

    @staticmethod
    def sentinel_to_nis_preparatory(sentinel_data_package_json_path: str, nis_file: str, log: str = None):
        """
        Read a Sentinel Data Package and elaborate an XLSX file with NIS commands and complementary information
        describing what is inside the Sentinel Package (enumeration of Scenarios, Regions, etc.) and a template of
        CORRESPONDENCE file as one of the Worksheets.

        :param sentinel_data_package_json_path: The full path of the JSON file describing a Sentinel data package
        :param nis_file: Full path of the output XLSX NIS formatted file
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        sentinel_to_prep_file(sentinel_data_package_json_path, nis_file)

    @staticmethod
    def list_lcia_indicators(nis_file: str, output_file: str,
                             compartment: bool = False, subcompartment: bool = False, horizon: bool = False,
                             log: str = None):
        """
        Parse the NIS file and extract a list of LCIA indicators

        :param nis_file: Full path of the NIS file (local or Google Drive, .XLSX format)
        :param output_file: Full path of the output .XLSX file
        :param compartment: Include compartment in the generated LCIAMethod calls
        :param subcompartment: Include subcompartment in the generated LCIAMethod calls
        :param horizon: Include horizon in the generated LCIAMethod calls
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :return:
        """
        set_log_level_from_cli_param(log)
        list_lcia_indicators(nis_file, output_file, compartment, subcompartment, horizon)

    @staticmethod
    def enviro(cfg_file_path,
               just_prepare_base: bool = False,
               fragments_list_file: bool = False,
               first_fragment: int = 0,
               n_fragments: int = 0,
               max_lci_interfaces: int = 0,
               keep_min_fragment_files: bool = True,
               generate_nis_base_file: bool = False,
               generate_full_fragment_files: bool = False,
               generate_interface_results: bool = False,
               generate_indicators: bool = False,
               n_cpus: int = 1,
               log: str = None,
               ):
        """
        The main function of the package, reads the contents of the configuration file, which can be either a JSON or YAML
        file. Following an example of YAML file contents:
            nis_file_location: "https://docs.google.com/spreadsheets/d/12AlJ0tdu2b-cfalNzLqFYfiC-hdDlIv1M1pTE3AfSWY/edit#gid=1986791705"
            correspondence_files_path: ""
            simulation_type: sentinel
            simulation_files_path: ""
            output_directory: ""

        (you can copy and paste the example in an empty text file as reference, and ".yaml" as extension)

        :param cfg_file_path: Path of the configuration file in YAML format
        :param n_fragments: number of fragments to process, 0 for "all"
        :param first_fragment: Index of the first fragment to be processed. To obtain an ordered list of fragments, execute first "enbios enviro" with "--just-prepare-base --fragments-list-file" options
        :param generate_nis_base_file: True if the Base file should be generated (once) for testing purposes
        :param generate_full_fragment_files: True to generate a full NIS formatted XLSX file for each fragment
        :param generate_interface_results: True to generate a CSV with values at interfaces for each fragment
        :param generate_indicators: True to generate a CSV with indicators for each fragment
        :param fragments_list_file: True to generate a CSV with the list of fragments
        :param keep_min_fragment_files: If True, do not delete minimal NIS files submitted to NIS to compute indicators
        :param max_lci_interfaces: Max number of LCI interfaces to consider. 0 for all (default 0)
        :param n_cpus: Number of CPUs of the local computer used to perform the process (default 1, sequential; 0 to find a good value for the computer automatically)
        :param log: Set log level to one of: Error (E, Err), Debug (D), Warning (W, Warn), Info (I), Off, Critical (Fatal)
        :param just_prepare_base: True to only prepare (download, parse and execute, then cache; but not solve) Base file and exit
        :return:
        """
        # locale.setlocale(locale.LC_ALL, 'en_US')
        set_log_level_from_cli_param(log)
        t = Enviro()
        t.set_cfg_file_path(cfg_file_path)

        t.compute_indicators_from_base_and_simulation(n_fragments,
                                                      first_fragment,
                                                      generate_nis_base_file,
                                                      generate_full_fragment_files,
                                                      generate_interface_results,
                                                      keep_min_fragment_files,
                                                      generate_indicators,
                                                      fragments_list_file,
                                                      max_lci_interfaces,
                                                      n_cpus,
                                                      just_prepare_base)

    @staticmethod
    def visualize(cfg_file_path, open_browser=False, log: str = None):
        """
        Prepare a Dash server enabling basic visualization of results
            - scenario
            - region
            - year
            - processor
            - [carrier]
            - indicator

        :param cfg_file_path:
        :param open_browser: If True, open the link in the browser
        :param log:
        :return:
        """
        # Read data
        print("Starting visualization server")
        cfg_file_path = os.path.realpath(cfg_file_path)
        print("Parsing configuration")
        cfg = read_parse_configuration(cfg_file_path)
        base_dir = cfg["output_directory"]
        if not os.path.isabs(base_dir):
            base_dir = os.path.join(os.path.dirname(cfg_file_path), base_dir)

        print("Creating Dash app")
        app = create_dashboard_app(base_dir)

        # Launch Dash dashboard server
        host = "localhost"
        port = 8050
        address = f"http://{host}:{port}"
        print(f"Launching server. Please, open {address}. Ctrl+C to Stop.")
        if open_browser:
            webbrowser.open(address)
        app.run_server(host=host, port=port, use_reloader=False)


def main():
    import platform
    os.environ["PAGER"] = "cat" if platform.system().lower() != "windows" else "-"
    fire.Fire(Enbios)


if __name__ == '__main__':
    main()



import functools
import logging
import operator
import os
import sys
import time
from multiprocessing import Pool, cpu_count
import itertools

import pandas as pd
from typing import Tuple, Dict, Set, List, Optional
from NamedAtomicLock import NamedAtomicLock
from friendly_data.converters import from_df
from nexinfosys.bin.cli_script import get_valid_name, get_file_url, prepare_base_state, print_issues
from nexinfosys.command_generators import IType, Issue
from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression

from nexinfosys.common.helper import PartialRetrievalDictionary, any_error_issue, create_dictionary
from nexinfosys.embedded_nis import NIS
from nexinfosys.model_services import State
from nexinfosys.models.musiasem_concepts import Processor, ProcessorsRelationPartOfObservation, Indicator, \
    FactorQuantitativeObservation
from nexinfosys.serialization import deserialize_state

from enbios.common.helper import generate_workbook, list_to_dataframe, get_scenario_name
from enbios.input import Simulation
from enbios.input.lci import LCIIndex
from enbios.input.simulators.sentinel import SentinelSimulation
from enbios.input.simulators.single_csv import AllInACSV
from enbios.model import SimStructuralProcessorAttributes, g_default_subtech
from enbios.processing import read_parse_configuration, read_submit_solve_nis_file


#####################################################
# MAIN ENTRY POINT  #################################
#####################################################
_idx_cols = ["region", "scenario", "subscenario", "technology", "subtechnology", "model", "carrier",
             "year", "time", "timestep", "unit", "variable", "description"]


def parallelizable_process_fragment(param: Tuple[str,  # Fragment label
                                                 Dict[str, str],  # Fragment dict
                                                 Dict[str, Set[str]],  #
                                                 List[SimStructuralProcessorAttributes]],
                                    s_state: bytes,
                                    output_dir: str,
                                    development_nis_file: str,
                                    generate_nis_fragment_file: bool,
                                    generate_interface_results: bool,
                                    generate_indicators: bool,
                                    max_lci_interfaces: int,
                                    keep_fragment_file: bool
                                    ):
    """
    Prepares a NIS file from inputs and submits it to NIS for accounting and calculation of indicators

    :param param: A Tuple with the information to drive the process
    :param s_state: A "bytes" with Serialized state (deserialized inside)
    :param output_dir: Outputs directory
    :param development_nis_file: URL of a NIS file that would go after state and before the fragment,
                                parsed everytime (for experiments)
    :param generate_nis_fragment_file: True to generate an expanded NIS file for the fragment
    :param generate_interface_results: True to generate a NIS file with the values of interfaces after solving
    :param generate_indicators: True to generate a file with all indicators
    :param max_lci_interfaces: If >0, cut the LCI interfaces used to that number
    :param keep_fragment_file: If True, keep the minimal fragment NIS file
    :return:
    """

    def write_outputs(outputs):
        """
        Write results of a submission to output directory

        :param outputs: Dictionary with the possible outputs
        :return:
        """
        nis_idempotent_file = outputs.get("idempotent_nis", None)
        df_indicators = outputs.get("indicators", pd.DataFrame())
        df_global_indicators = outputs.get("global_indicators", pd.DataFrame())
        df_interfaces = outputs.get("interfaces", pd.DataFrame())
        df_iamc_indicators = outputs.get("iamc_indicators", pd.DataFrame())
        print("Writing results ...")

        def append_fragment_to_file(file_name, df):
            lock = NamedAtomicLock("enbios-lock")
            lock.acquire()
            try:
                full_file_name = os.path.join(output_dir, file_name)
                if not os.path.isfile(full_file_name):
                    df.to_csv(full_file_name, index=False)
                else:
                    df.to_csv(full_file_name, index=False, mode='a', header=False)
            finally:
                lock.release()

        # Main result: "indicators.csv" file (aggregates all fragments)
        if df_indicators is not None:
            append_fragment_to_file("indicators.csv", df_indicators)

            # Write a separate indicators.csv for the fragment
            if generate_indicators:
                csv_name = os.path.join(output_dir, f"fragment_indicators{get_valid_name(str(p_key))}.csv")
                if not df_indicators.empty:
                    df_indicators.to_csv(csv_name, index=False)
                else:
                    with open(csv_name, "wt") as f:
                        f.write("Could not obtain indicator values (??)")

        # "system_global_indicators.csv" file (aggregates all fragments)
        if df_global_indicators is not None:
            append_fragment_to_file("system_global_indicators.csv", df_global_indicators)

        if df_iamc_indicators is not None:
            append_fragment_to_file("iamc_indicators.csv", df_iamc_indicators)

        # Write NIS of the fragment just processed
        if generate_nis_fragment_file and nis_idempotent_file is not None:
            fragment_file_name = os.path.join(output_dir, f"full_fragment{get_valid_name(str(p_key))}.xlsx")
            with open(fragment_file_name, "wb") as f:
                f.write(nis_idempotent_file)

        # Write Dataset with values for each interface as calculated by NIS solver, for the fragment
        if generate_interface_results and df_interfaces is not None:
            csv_name = os.path.join(output_dir, f"fragment_interfaces{get_valid_name(str(p_key))}.csv")
            if not df_interfaces.empty:
                df_interfaces.to_csv(csv_name, index=False)
            else:
                with open(csv_name, "wt") as f:
                    f.write("Could not obtain interface values (??)")

    # Starts here
    frag_label, p_key, f_metadata, f_processors = param  # Unpack "param"
    print(f"Fragment processing ...")
    start = time.time()  # Time execution

    # Call main function
    outputs = process_fragment(s_state, p_key,
                               f_metadata, f_processors,
                               output_dir,
                               development_nis_file,
                               max_lci_interfaces,
                               keep_fragment_file,
                               generate_nis_fragment_file,
                               generate_interface_results)
    end = time.time()  # Stop timing
    print(f"Fragment processed in {end - start} seconds ---------------------------------------")
    write_outputs(outputs)
    start = time.time()  # Time also output writing
    print(f"Fragment outputs written in {start-end} seconds to output dir: {output_dir}")


class Enviro:
    def __init__(self):
        self._cfg_file_path = None
        self._cfg = None
        self._simulation_files_path = None

    def set_cfg_file_path(self, cfg_file_path):
        self._cfg = read_parse_configuration(cfg_file_path)
        self._cfg_file_path = os.path.realpath(cfg_file_path) if isinstance(cfg_file_path, str) else None
        if "simulation_files_path" in self._cfg:
            self._simulation_files_path = self._cfg["simulation_files_path"]

    def _get_simulation(self) -> Simulation:
        # Simulation
        simulation = None
        if self._cfg["simulation_type"].lower() == "sentinel":
            simulation = SentinelSimulation(self._simulation_files_path)
        elif self._cfg["simulation_type"].lower() == "single_csv":
            simulation = AllInACSV(self._simulation_files_path)

        # elif self._cfg["simulation_type"].lower() == "calliope":
        #     simulation = CalliopeSimulation(self._simulation_files_path)
        return simulation

    def _prepare_process(self) -> Tuple[NIS, LCIIndex, Simulation]:
        # Simulation
        simulation = self._get_simulation()
        # MuSIASEM (NIS)
        nis, issues = read_submit_solve_nis_file(self._cfg["nis_file_location"], state=None, solve=False)
        # LCI index
        lci_data_index = LCIIndex(self._cfg["lci_data_locations"])

        return nis, lci_data_index, simulation

    def generate_matcher_templates(self, combine_countries=False) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Generate auxiliary DataFrames to help in the elaboration of:
          - correspondence file
          - block types file
        :param combine_countries:
        :return: A tuple with 4 pd.DataFrames
        """
        # Construct auxiliary models
        nis, lci_data_index, simulation = self._prepare_process()
        # TODO Build a DataFrame with a list of LCI title, to file code
        #   name,title,file
        lci_df = pd.DataFrame()
        # TODO Build a DataFrame with a list of BareProcessors
        #   processor,parent
        nis_df = pd.DataFrame()
        # TODO Build DataFrame "correspondence"
        #   name,match_target_type,match_target,weight,match_conditions
        #   OK   "lca"             ---          1      <generate>
        #  if combine_countries:
        #   for each <name> generate a line per country and a line without country
        correspondence_df = pd.DataFrame()
        # TODO Build DataFrame "block_types"
        #   name,type
        #   OK   <from an existing database of tech names to Calliope tech types>
        block_types_df = pd.DataFrame()
        return lci_df, nis_df, correspondence_df, block_types_df

    def _read_simulation_fragments(self, split_by_region=True, split_by_scenario=True, split_by_period=True, default_time=None):

        """
        Read simulation. Find fragments and then start an iteration on them
        Each fragment is made of the "processors" and their "interfaces"
        It is here where basic processors are built. Correspondence file has tech but it could also
        have carrier and region

        :return:
        """
        # Create simulation
        simulation = self._get_simulation()
        # Read Simulation, returning:
        # prd: PartialRetrievalDictionary
        # scenarios: list of scenarios
        # regions: list of regions
        # times: list of time periods
        # carriers: list of carriers
        # units: list of units
        # col_types: list of value (not index) fields
        # ctc: list of pairs (tech, region, carrier)
        prd, scenarios, regions, times, techs, carriers, units, col_types, ctc = simulation.read("", default_time)

        partition_lists = []
        mandatory_attributes = ["carrier", "technology"]
        if split_by_region and len(regions) > 0:
            partition_lists.append([("_g", r) for r in regions])
            mandatory_attributes.append("region")
        if split_by_period and len(times) > 0:
            partition_lists.append([("_d", t) for t in times])
            mandatory_attributes.append("time")
        # TODO Spores
        if split_by_scenario and len(scenarios) > 0:
            partition_lists.append([("_s", s) for s in scenarios])
            mandatory_attributes.append("scenario")

        for i, partition in enumerate(sorted(list(itertools.product(*partition_lists)))):
            partial_key = {t[0]: t[1] for t in partition}
            procs = prd.get(partial_key)
            # Sweep "procs" and update periods, regions, scenarios, models and carriers
            md = dict(periods=set(), regions=set(), scenarios=set(), models=set(), carriers=set(), techs=set(), flows=set())
            accountable_procs = []
            for p in procs:
                any_mandatory_not_found = False
                for mandatory in mandatory_attributes:
                    if mandatory not in p.attrs:
                        # print(f"Processor {p.attrs} did not define '{mandatory}'. Skipping.")
                        any_mandatory_not_found = True
                        break
                if any_mandatory_not_found:
                    continue
                accountable_procs.append(p)
                if "region" in p.attrs:
                    md["regions"].add(p.attrs["region"])
                if "scenario" in p.attrs:
                    md["scenarios"].add(p.attrs["scenario"])
                if "technology" in p.attrs:
                    md["techs"].add(p.attrs["technology"])
                if "time" in p.attrs:
                    md["periods"].add(p.attrs["time"])
                if "model" in p.attrs:
                    md["models"].add(p.attrs["model"])
                if "carrier" in p.attrs:
                    md["carriers"].add(p.attrs["carrier"])
                tmp = set(p.attrs.keys()).difference(_idx_cols)
                md["flows"].update(tmp)
            # Return 4 items:
            #  - a fragment label, to sort fragments
            #  - an equivalent (to the label) dict
            #  - fragment metadata
            #  - the Processors (the data to process!)
            # print(':'.join([f"{k}{v}" for k, v in partial_key.items()]))

            yield ':'.join([f"{k}{v}" for k, v in partial_key.items()]), partial_key, md, accountable_procs

    def compute_indicators_from_base_and_simulation(self,
                                                    n_fragments: int = 0,
                                                    first_fragment: int = 0,
                                                    generate_nis_base_file: bool = False,
                                                    generate_nis_fragment_file: bool = False,
                                                    generate_interface_results: bool = False,
                                                    keep_fragment_file: bool = True,
                                                    generate_indicators: bool = False,
                                                    fragments_list_file: bool = False,
                                                    max_lci_interfaces: int = 0,
                                                    n_cpus: int = 1,
                                                    just_prepare_base: bool = False):
        """
        MAIN entry point of current ENVIRO
        Previously, a Base NIS must have been prepared, see @_prepare_base

        :param n_fragments: number of fragments to process, 0 for "all"
        :param first_fragment: Index of the first fragment to be processed. To obtain an ordered list of fragments, execute first "enbios enviro" with --fragments-list-file option
        :param generate_nis_base_file: True if the Base file should be generated (once) for testing purposes
        :param generate_nis_fragment_file: True to generate a full NIS formatted XLSX file for each fragment
        :param generate_interface_results: True to generate a CSV with values at interfaces for each fragment
        :param generate_indicators: True to generate a CSV with indicators for each fragment
        :param fragments_list_file: True to generate a CSV with the list of fragments
        :param max_lci_interfaces: Max number of LCI interfaces to consider. 0 for all (default 0)
        :param n_cpus: Number of CPUs of the local computer used to perform the process
        :param just_prepare_base: True to only prepare (download, parse and execute, then cache; but not solve) Base file and exit
        :param keep_fragment_file: If True, do not remove the minimal NIS file generated and submitted to NIS
        :return:
        """

        output_dir = self._cfg["output_directory"]
        if not os.path.isabs(output_dir):
            output_dir = os.path.join(os.path.dirname(self._cfg_file_path), output_dir)

        print(f"output_dir: {output_dir}")
        os.makedirs(output_dir, exist_ok=True)

        # Prepare Base
        state, serial_state, issues = prepare_base_state(self._cfg["nis_file_location"], False, output_dir)

        if not state:
            logging.debug(f"NIS BASE PREPARATION FAILED")
            print_issues("NIS base preparation", self._cfg["nis_file_location"], issues,
                         f"Base not ready due to errors, exiting. Please check the issues")
            sys.exit(1)

        if just_prepare_base and not fragments_list_file:
            print("Base processed and cached, exiting because 'just_prepare_base == True'")
            sys.exit(1)

        if "development_nis_file_location" in self._cfg:
            development_nis_file = self._cfg["development_nis_file_location"]
        else:
            development_nis_file = None

        # [Write Base as a full NIS file]
        if generate_nis_base_file:
            issues, new_state, results = run_nis_for_results(None, None, state, ["model"])
            nis_file = results[0]
            with open(os.path.join(output_dir, "nis_base.idempotent.xlsx"), "wb") as f:
                f.write(nis_file)

        # Read simulation AND Split it in decoupled fragments
        default_time = self._cfg.get("simulation_default_time")
        fragments = sorted([_ for _ in self._read_simulation_fragments(default_time=default_time)],
                           key=operator.itemgetter(0))
        fragments = [_ for _ in fragments if len(_[3]) > 0]

        # Produce file with an enumeration of fragments
        if fragments_list_file:
            possible_columns = dict(_g="Regions", _d="Periods", _s="Scenarios")
            used_columns = []
            used_columns_set = set()
            for fragment in fragments:
                for key in fragment[1].keys():
                    if key not in used_columns_set:
                        used_columns.append(key)
                        used_columns_set.add(key)
            s = [", ".join([possible_columns[_] for _ in used_columns])]
            for fragment in fragments:
                s.append(", ".join([f"{fragment[1].get(_, '')}" for _ in used_columns]))
            with open(os.path.join(output_dir, "fragments_list.csv"), "w") as f:
                f.write("\n".join(s))

        logging.debug(f"Simulation read, and split in {len(fragments)} fragments")

        if just_prepare_base:
            print("Base processed and cached, exiting because 'just_prepare_base == True'")
            sys.exit(1)

        # Remove "indicators.csv" and "iamc_indicators if they exist
        for f_name in ["indicators.csv", "iamc_indicators.csv"]:
            f_path = os.path.join(output_dir, f_name)
            if os.path.isfile(f_path):
                os.remove(f_path)

        # Possibly reduce the list of fragments to process
        if n_fragments > 0 or first_fragment > 0:
            if n_fragments == 0:
                fragments = fragments[first_fragment:]
            else:
                fragments = fragments[first_fragment:first_fragment + n_fragments]
            logging.debug(f"{n_fragments} fragment{'s' if n_fragments > 1 else ''} will be processed, starting from fragment {first_fragment}")

        # PROCESS FRAGMENTS - Sequential or embarrassingly parallel
        # If 0 -> find an appropriate number of CPUs to use. int(80% of # CPUs) if more than 4; 2 for 4; 1 for < 4
        if n_cpus == 0:
            n_cpus = int(0.8*cpu_count()) if cpu_count() > 4 else 2 if cpu_count() == 4 else 1

        if n_cpus == 1:
            for i, (frag_label, partial_key, frag_metadata, frag_processors) in enumerate(fragments):
                # fragment_metadata: dict with regions, years, scenarios in the fragment
                # fragment_processors: list of processors with their attributes which will be interfaces
                # print(f"{partial_key}: {len(frag_processors)}")
                parallelizable_process_fragment((frag_label, partial_key, frag_metadata, frag_processors), serial_state,
                                                output_dir,
                                                development_nis_file,
                                                generate_nis_fragment_file, generate_interface_results,
                                                generate_indicators, max_lci_interfaces, keep_fragment_file)
        else:
            p = Pool(n_cpus)
            p.map(functools.partial(parallelizable_process_fragment,
                                    s_state=serial_state,
                                    output_dir=output_dir,
                                    development_nis_file=development_nis_file,
                                    generate_nis_fragment_file=generate_nis_fragment_file,
                                    generate_interface_results=generate_interface_results,
                                    generate_indicators=generate_indicators,
                                    max_lci_interfaces=max_lci_interfaces, keep_fragment_file=keep_fragment_file),
                  fragments)


def run_nis_for_results(nis_file_name: Optional[str],
                        development_nis_file: Optional[str],
                        _state: State,
                        requested_outputs: List[Tuple] = None):
    """
    Call to NIS and return results

    :param nis_file_name: File to process
    :param development_nis_file: URL of a NIS file that would go after state and before the fragment,
                                parsed everytime (for experiments)
    :param _state: Initial State assumed by the file to process
    :param requested_outputs:
    :return: Issues, new State, and either: a) the requested outputs (None entry if error) or
                                            b) a list of available outputs
    """
    outputs = {
        "flow_graph_solution": ("dataset", "flow_graph_solution", "csv"),
        "flow_graph_solution_indicators": ("dataset", "flow_graph_solution_indicators", "csv"),
        "flow_graph_solution_edges": ("dataset", "flow_graph_solution_edges", "csv"),
        "flow_graph_global_indicators": ("dataset", "flow_graph_global_indicators", "csv"),
        "flow_graph_solution_benchmarks": ("dataset", "flow_graph_solution_benchmarks", "csv"),
        "flow_graph_solution_global_benchmarks": ("dataset", "flow_graph_solution_global_benchmarks", "csv"),
        "benchmarks_and_stakeholders": ("dataset", "benchmarks_and_stakeholders", "csv"),
        "model": ("model", "Model.xlsx"),
        "interfaces_graph": ("graph", "interfaces_graph.visjs"),
        "processors_graph": ("graph", "processors_graph.visjs"),
        "processors_geolayer": ("geolayer", "processors_geolayer.geojson"),
    }
    nis = NIS()
    nis.open_session(True, _state)
    if nis_file_name:
        if development_nis_file:
            nis.load_workbook(development_nis_file)
        nis.load_workbook(get_file_url(nis_file_name))
        issues = nis.submit_and_solve()
        error = any_error_issue(issues)
        new_state = nis.get_state()
    elif isinstance(_state, State):
        # nis.append_command("Ignore me", list_to_dataframe([["dummy"], ["dummy"]]))
        issues = []
        error = False
        new_state = _state
    else:
        issues = [Issue(itype=IType.ERROR,
                        description=f"At least NIS file or initial State must be specified",
                        location=None)]
        error = True
        new_state = None
    if not error:
        if requested_outputs:
            _ = []
            for r in requested_outputs:
                if isinstance(r, str):
                    if r in outputs:
                        _.append(outputs[r])
                    else:
                        issues.append(Issue(itype=IType.ERROR,
                                            description=f"Dataset '{r}' not registered",
                                            location=None))
                elif isinstance(r, tuple):
                    _.append(r)
            _ = [r[0] if r[2] else None for r in nis.get_results(_)]
        else:
            _ = nis.query_available_datasets()
    else:
        _ = []

    nis.close_session()

    return issues, new_state, _


def process_fragment(base_serial_state: bytes,
                     partial_key: Dict[str, str],
                     fragment_metadata: Dict[str, Set[str]],
                     fragment_processors: List[SimStructuralProcessorAttributes],
                     output_directory: str,
                     development_nis_file: str,
                     max_lci_interfaces: int,
                     keep_fragment_file: bool,
                     generate_nis_fragment_file: bool,
                     generate_interface_results: bool
                     ):
    """
    :param base_serial_state: A "bytes" with Serialized state (deserialized inside)
    :param partial_key: A dictionary describing the fragment dimensions
    :param fragment_metadata: A dictionary with sets of dimensions inside the fragment
    :param fragment_processors: A list of the structural processor attributes (read from simulation output)
                                inside the fragment.
    :param output_directory: Outputs directory
    :param development_nis_file: URL of a NIS file that would go after state and before the fragment,
                                parsed everytime (for experiments)
    :param max_lci_interfaces: If >0, cut the LCI interfaces used to that number
    :param keep_fragment_file: If True, do not remove the minimal fragment file submitted to NIS
    :param generate_nis_fragment_file: If True generate the idempotent NIS model
    :param generate_interface_results: If True generate the interface results (used to do the indicator calculations)

    Different assembly options
     * Dendrogram functional processors: a clone per region / a fragment per region (so no need to clone)
    Spold:
     * Structural processors: currently there is a list of structural (not-accounted) processors attached to the
     last level of functional processors, with a .spold file associated with them.
     When a simulation processor is found, associate it with one of these processors (equal "tech" name)
     * If there is a .spold file associated to the structural processor associated to the simulation, use it
     * If there is no .spold file linked to this processor and the "parent processor" has one:
       - attach the simulation processor to the parent processor
       - use the spold file of the parent

     Something like:
          if spold(Sim) is None
            if spold(P(Sim)) -> spold(Sim) = spold(P(Sim))
            else ERR
          else OK
          INSERT Sim interfaces
          output_interface_name <- name of the output interface from interfaces in the simulation processor
          for spold_proc in spold(Sim)
            for each interface in spold_proc
              INSERT interface, make them relative_to "output_interface_name"
          COPY all interfaces of spold(Sim) to Sim (insert "interfaces") and make them relative_to the output interface

          if skip_aux_structural AND is_structural(P(Sim)) AND is_not_accounted(P(Sim)) -> P(Sim) = P(P(Sim))
    """

    # Declare (InterfaceType) the original (from simulation) name even if it is renamed (EcoinventCarrierName) and
    # duplicate the Interface with the original and the changed name
    write_original_interface = True

    def _get_processors_by_type():
        """
        Returns two sets of Processors:
          - "Functional without Interfaces": MuSIASEM dendrogram (without data, so no top-down would be possible)
          - "Structural not Accounted": LCI processors and Simulation processors

        NOTE: these two sets are not a partition of all Processors: we may have "Functional that may have Interfaces"
              and "Structural which are Accounted"

        The implicit input is the state in "prd" (PartialRetrievalDictionary)
        :return: The two sets
        """
        procs = prd.get(Processor.partial_key())
        musiasem_dendrogram_procs = {}
        structural_procs = {}
        for proc in procs:
            n_interfaces = len(proc.factors)
            accounted = proc.instance_or_archetype.lower() == "instance"
            functional = proc.functional_or_structural.lower() == "functional"
            if accounted and n_interfaces == 0:
                for h_name in proc.full_hierarchy_names(prd):
                    musiasem_dendrogram_procs[h_name] = proc
            elif not functional and not accounted:
                for h_name in proc.full_hierarchy_names(prd):
                    structural_procs[h_name] = proc
            else:
                print(f"{proc.full_hierarchy_names(prd)[0]} is neither MuSIASEM base nor LCI base "
                      f"(functional: {functional}; accounted: {accounted}; n_interfaces: {n_interfaces}.")
        return musiasem_dendrogram_procs, structural_procs

    def _find_parents(registry: PartialRetrievalDictionary, base_procs: Dict[str, Processor], p: Processor):
        """
        Find names of parent Processors of "p" (not all ancestors, just parents).

        :param registry: Registry of NIS objects
        :param base_procs: Dict of Processor in the dendrogram
        :param p: name of a simulation process, which should be enumerated in the previous Dict
        :return: Dict of parents, {processor_name: Processor}. None if "p" is not found
        """
        return [rel.parent_processor
                for rel in registry.get(ProcessorsRelationPartOfObservation.partial_key(child=p))]

        # Find the Processor object
        # SimProc(tech, carrier) ->(child-of)-> FunctMuSProc(ParentProc)
        # SimProc(tech, carrier) as StrucMuSProc(tech, carrier2)

    def _find_lci_and_tech_processors(reg: PartialRetrievalDictionary, structural_procs, base_procs, p) -> Dict[Processor, List]:
        """
        Find LCI Processor(s) which are assimilated to "p"
          - Processor "p" should exist, as Structural/not-Accounted, with a Parent, [with no Interfaces?]
          - Ecospold file potentially associated
          - if "p" does not have an Ecospold file associated, take parent's Ecospold file
          - Find an "LCI processor": Structural/not-Accounted, NO Parent, with the same Ecospold

        :param reg: Registry of NIS objects
        :param structural_procs: Dict of Structural not Accounted Processors, which includes LCI processors but
                                 also Simulation processors
        :param base_procs: Dict of Dendrogram Processors
        :param p: Name of the target processor
        :return: List of matching LCI processors, None if "p" is not found; tech processor tech output variable name, desired output name (carrier name)
        """
        def _find_reference_lci_processor(target: Processor, lci_procs: Dict[str, Processor]):
            lci_proc = None
            ecospold_file = target.attributes.get("EcoinventFilename", "")
            if ecospold_file != "":
                for proc_name, proc in lci_procs.items():
                    if "." not in proc_name:
                        ecospold_file_2 = proc.attributes.get("EcoinventFilename", "")
                        if len(ecospold_file) > 10 and len(ecospold_file_2) > 10 and \
                           (ecospold_file.lower() in ecospold_file_2.lower() or
                           ecospold_file_2.lower() in ecospold_file.lower()):
                            lci_proc = proc
                            break
            return lci_proc

        # Find structural not accounted matching the sim processor (match tech name and proc carrier name)
        targets: Dict[Processor, List] = {}
        tech = p.attrs["technology"].lower()
        carrier_ = p.attrs["carrier"].lower()  # Mandatory
        for proc_name, proc in structural_procs.items():
            last = proc_name.split(".")[-1]
            if tech == last.lower():
                if carrier_ != "":
                    carrier_field = ""
                    if "SimulationCarrier" in proc.attributes:
                        carrier_field = "SimulationCarrier"
                    elif "EcoinventCarrierName" in proc.attributes:
                        carrier_field = "EcoinventCarrierName"  # It should never enter here
                    if proc.attributes.get(carrier_field, "").lower() == carrier_:
                        targets[proc] = []
                else:
                    targets[proc] = []  # If "tech" matches and "carrier_" is ""

        for target in targets.keys():
            lci_proc = _find_reference_lci_processor(target, structural_procs)
            if lci_proc is None:
                # Parent
                s1 = target.full_hierarchy_names(reg)[0].rsplit(".", 1)[0]
                for proc_name, proc in base_procs.items():
                    if s1 == proc_name:
                        lci_proc = _find_reference_lci_processor(proc, structural_procs)
                        targets[target].append(lci_proc)
                        break
            else:
                targets[target].append(lci_proc)

        return targets

    def _generate_fragment_nis_file(clone_processors, cloners, processors, interfaces, variables, scenarios):
        """
        Once model has been assembled in memory, generate a NIS file
        This file can be later submmited for it to be accounted and indicators calculated

        :param clone_processors:
        :param cloners:
        :param processors:
        :param interfaces:
        :param variables:
        :param scenarios:
        :return:
        """
        cmds = []
        # ProblemStatement
        lst = [["Scenario", "Parameter", "Value", "Description"]]
        for s in scenarios:
            # Scenario, Parameter, Value, Description
            scen = get_scenario_name("s", s)
            observer = get_scenario_name("o", s)
            lst.append([scen, "NISSolverObserversPriority", observer, f"Scenario {scen}, observer {observer}"])
        cmds.append(("ProblemStatement sim fragment", list_to_dataframe(lst)))
        # RefProvenance
        lst = [["RefID", "ProvenanceFileURL", "AgentType", "Agent", "Activities", "Entities"]]
        for s in scenarios:
            observer = get_scenario_name("o", s)
            scen = get_scenario_name("s", s)
            lst.append([observer, "", "Software", f"Sentinel simulation observer for scenario {scen}", "WP4", ""])
        cmds.append(("RefProvenance sim fragment", list_to_dataframe(lst)))
        # InterfaceTypes
        lst = [["InterfaceTypeHierarchy", "InterfaceType", "Sphere", "RoegenType", "ParentInterfaceType", "Formula",
                "Description", "Unit", "OppositeSubsystemType", "Attributes"]]
        for v_name, attrs in variables.items():
            if not attrs["main_flow"]:
                lst.append(["Sentinel", v_name, "Technosphere", "Flow", "", "", "", attrs["unit"], "", ""])
        cmds.append(("InterfaceTypes sim fragment", list_to_dataframe(lst)))

        if clone:
            cmds.append(("BareProcessors regions", list_to_dataframe(clone_processors)))
            cmds.append(("ProcessorScalings", list_to_dataframe(cloners)))
        cmds.append(("BareProcessors sim fragment", list_to_dataframe(processors)))
        cmds.append(("Interfaces sim fragment", list_to_dataframe(interfaces)))
        s = generate_workbook(cmds)
        if s:
            frag_file_name = os.path.join(output_directory, f"fragment_{get_valid_name(str(partial_key))}.xlsx")
            with open(frag_file_name, "wb") as f:
                f.write(s)
            return frag_file_name
        else:
            return None

    def is_main_flow(flow_name, output_flow="flow_out_sum"):  # TODO Depends on simulation
        return flow_name in [output_flow]

    def get_flow_orientation(flow_name):  # TODO Depends on simulation
        return "Output" if "out" in flow_name else "Input"

    def get_flow_unit(flow_name):  # TODO Depends on simulation
        if flow_name.lower() in ["flow_out", "flow_in", "flow_out_sum", "flow_in_sum"]:
            return "kWh"
        elif flow_name.lower() in ["capacity_factor"]:
            return "kW"
        else:
            return "dimensionless"

    def _interface_used_in_some_indicator(iface):
        # TODO Implement
        return True

    def _add_clone(cloners_list, clone_processors_list, regions, base_musiasem_procs):
        cloners_list.append(
            ["InvokingProcessor", "RequestedProcessor", "ScalingType", "InvokingInterface", "RequestedInterface",
             "Scale"])
        clone_processors_list.append(
            ["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
             "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong",
             "Attributes",
             "@SimulationName", "@Region"])
        # For each region, create a processor in "clone_processors" and an entry in "cloners" hanging the top
        # Processors in the dendrogram into the region processor
        for reg in regions:
            clone_processors_list.append(["", reg, "", "", reg, "Functional", "Yes", "",
                                         f"Country level processor, region '{reg}'", "", "", "", "", reg, reg])
            considered = set()
            for p_name, proc in base_musiasem_procs.items():
                if len(p_name.split(".")) > 1 and p_sim not in considered:
                    considered.add(proc)
                    # InvokingInterface, RequestedInterface, Scale are mandatory; specify something here
                    cloners_list.append((reg, p_name, "Clone", "", "", ""))

    def _get_processor_name(p):
        return f"{p.attrs['technology']}_{p.attrs['carrier']}"

    def _split_processor_name(p, carriers, techs):
        """
        Split a processor name into technology and carrier, in accordance with "_get_processor_name"

        :param p:
        :param carriers:
        :param techs:
        :return:
        """
        # Metadata contains technologies and carriers, which are code lists
        r_ = p.split(".")
        if r_[-1].endswith(tuple(carriers)) and r_[-1].startswith(tuple(techs)):
            # Find carrier and split
            # (carriers must be sorted by length, to avoid cases like this. If the carrier is "syn_diesel" and
            #  we have ["diesel", "syn_diesel"], it will fail. But if "carriers" is sorted by length, it will
            #  split "r_" properly)
            for c in carriers:
                if r_[-1].endswith(c):
                    nc = len(c)
                    return f"{'.'.join(r_[:-1])}.{r_[-1][:-nc-1]}", r_[-1][-nc:]
        else:
            return p, ""

    def _add_processor(proc, matches, processors_list, already_added_processors_set, techs_used_in_regions, context):
        """
        Inserts elements in "processors_list", directly usable as worksheet rows in a BareProcessors command

        :param proc:
        :param matches:
        :param processors_list:
        :param already_added_processors_set:
        :param techs_used_in_regions:
        :param context:
        :return:
        """
        first_name = None
        return_simple = True
        # Add "Accounted Structural" Processors, hanging from each parent
        for parent_p in matches:
            # NEW PROCESSOR NAME: technology + carrier [+ region]
            p_name = _get_processor_name(proc)
            if p_name in techs_used_in_regions:
                return_simple = False
            else:
                techs_used_in_regions.add(p_name)
            parent_name = parent_p.name
            description = f"Simulation {proc.attrs['technology']} for {proc.attrs['carrier']} at region {region}, scenario {scenario}."
            original_name = f"{proc.attrs['technology']}"
            if first_name is None:
                first_name = f"{parent_name}.{p_name}"
                if (p_name, parent_name) not in already_added_processors_set:
                    processors_list.append(["", p_name, parent_name, "", region, "Structural", "Yes", "",
                                            description, "", "", "", "", original_name, region])
                    already_added_processors_set.add((p_name, parent_name))
            else:
                if (p_name, parent_name) not in already_added_processors_set:
                    processors_list.append(["", first_name, parent_name, "", region, "Structural", "Yes", "",
                                            description, "", "", "", "", original_name, region])
                    already_added_processors_set.add((p_name, parent_name))
        return first_name, parent_name  # p_name if return_simple else first_name, parent_name

    def _add_interfaces(proc, proc_name, proc_parent_name, lci_matches, matching_tech_proc, context) -> None:
        """
        Inserts elements in "interfaces", directly usable as worksheet rows in an Interfaces command

        :param proc:
        :param proc_name:
        :param proc_parent_name:
        :param lci_matches: Which LCI processors match
        :param matching_tech_proc: Which technology processor matches
        :param context: "time", "region", "scenario", ... in which interfaces are being added
        :return:
        """

        def get_tech_simulation_scaling_factor():
            _ = matching_tech_proc.attributes.get("SimulationScalingFactor", "1.0")
            ast = string_to_ast(arith_boolean_expression, _)
            state = State(context)
            issues = []
            res, variables = ast_evaluator(ast, state, None, issues)
            if res:
                return res
            else:
                raise ValueError(f"Could not evaluate {_}. Variables: {variables}. Issues: {issues}")

        relative_to = None
        tech_output_name = matching_tech_proc.attributes.get("SimulationVariable")
        tech_desired_output_name = matching_tech_proc.attributes.get("EcoinventCarrierName")
        tech_simulation_scaling_factor = get_tech_simulation_scaling_factor()
        tech_output_to_spold_factor = float(matching_tech_proc.attributes.get("SimulationToEcoinventFactor", "1.0"))
        observer = get_scenario_name("o", scenario)
        # Interfaces from MuSIASEM tech
        for i in matching_tech_proc.factors:
            for o in i.observations:
                if isinstance(o, FactorQuantitativeObservation):
                    interfaces.append([name, i.name, "", "", "", i.orientation, "", "", "", "", o.value,
                                       "", "", "", "", "", "", o.attributes["time"],
                                       o.observer.name if o.observer else "", "", ""])

        # Interfaces from simulation
        for i in set(proc.attrs.keys()).difference(_idx_cols):
            i_name = i
            input_i_name = i_name
            v = proc.attrs[i]
            if is_main_flow(i, tech_output_name):
                if tech_desired_output_name:
                    i_name = tech_desired_output_name
                v *= tech_output_to_spold_factor * tech_simulation_scaling_factor  # Change scale
                relative_to = i_name  # All LCI interfaces will be relative to main output
                orientation = "Output"
            else:
                orientation = get_flow_orientation(i)
            interfaces.append([proc_name, i_name, "", "", "", orientation, "", "", "", "", v,
                               "", "", "", "", "", "", time_, observer, "", ""])
            if write_original_interface and input_i_name != i_name:
                interfaces.append([proc_name, input_i_name, "", "", "", orientation, "", "", "", "", v,
                                   "", "", "", "", "", "", time_, observer, "", ""])
            if i_name not in variables:
                variables[i_name] = dict(main_flow=is_main_flow(i), orientation=orientation, unit=get_flow_unit(i))
            if write_original_interface and input_i_name not in variables:
                variables[input_i_name] = dict(main_flow=False, orientation=orientation, unit=get_flow_unit(i))

        if lci_matches[0] is None or len(lci_matches) == 0:
            print(f"{proc.attrs} does not have an LCI processor associated. "
                  f"Cannot provide information for indicators, skipped.")
            return

        # Interfaces from LCI
        # The following condition could happen if the fragment includes several Regions, Years and Scenarios.
        # LCI does not change -currently, regionalized LCI would be different-
        if (proc_name, proc_parent_name) not in already_added_lci:
            already_added_lci.add((proc_name, proc_parent_name))

            for cp in lci_matches:
                cont = 0
                for i in cp.factors:
                    if cont >= max_lci_interfaces:
                        break  # Artificial limit to reduce file size, for test purposes
                    if not _interface_used_in_some_indicator(
                            i.name):  # Economy of the model: avoid specifying interfaces not used later
                        continue
                    q = i.quantitative_observations[0]
                    if q.attributes.get("relative_to", None) is None:
                        continue
                    cont += 1
                    unit = q.attributes.get("unit")  # TODO unit / relative_to_unit
                    orientation = i.attributes.get("orientation", "")
                    v = q.value
                    # TODO Static: does not depend on scenario or time, avoid inserting repeatedly
                    interfaces.append([proc_name, i.name, "", "", "", orientation, "", "", "", "", v,
                                       "", relative_to, "", "", "", "", "Year", "Ecoinvent", "", ""])

    def _prepare_iamc_dataframe_fragment(state: State, indicators: pd.DataFrame, iamc_codes: Dict[str, str], metadata):
        # self._simulation_files_path
        # TODO Write the indicators using Sentinel-friendly-data-IAMC format
        #  from friendly_data.io import dwim_file (Do What I Mean)
        #  from friendly_data.iamc import IAMconv
        #  from friendly_data.dpkg import pkgindex
        #  conf = dwim_file("conf.yaml")
        #  conf["indices"]
        #  idx = pkgindex.from_file("index.yaml")
        #  converter = IAMconv(idx, conf["indices"], basepath="<location>")
        #  conf2 = conf["indices"]
        #  conf2["year"] = 2050
        #  converter2 = IAMconv(idx, conf2, basepath="")
        #  converter.to_csv o to_df
        #  .
        #  Also, e-mail from Suvayu
        # try:
        #     res = from_df(df_indicators, output_dir, os.path.join(output_dir, f"indicators_fd.csv"))
        # except:
        #     traceback.print_exc()
        iamc_df = indicators.copy(True)
        # Scope
        iamc_df = iamc_df[iamc_df["Scope"] == "Internal"]
        iamc_df = iamc_df.drop(columns=["Scope"])
        # Model
        iamc_df["Model"] = "enbios[calliope]"
        # Scenario, Region, Value, Unit, Year
        iamc_df.rename(columns=dict(System="Region", Scenario="Scenario", Value="Value", Unit="Unit", Period="Year"),
                       inplace=True)

        def _get_iamc_code(row, prd, iamc_codes, carriers, techs):
            processor, indicator = row["Processor"], row["Indicator"]
            # First part of the code, from the processor
            if processor in iamc_codes:
                first_part = iamc_codes[processor]
            else:
                parts = _split_processor_name(processor, carriers, techs)
                p = prd.get(Processor.partial_key(parts[0]))
                if p and len(p) == 1:
                    first_part = p[0].attributes.get("iamccode", "")
                else:
                    first_part = ""
            # Second part of the code, from the indicator
            ind = prd.get(Indicator.partial_key(indicator))
            if ind and len(ind) == 1:
                # TODO It could also be that if "IAMCCodeSuffix" was not specified, the second part is left empty,
                #   so the row is skipped
                second_part = ind[0].attributes.get("iamccodesuffix", indicator)
            else:
                second_part = indicator
            # Combine both parts and return
            if first_part and second_part:
                return f"{first_part}|{second_part}"
            else:
                return ""

        def _get_indicator_unit(row, prd):
            indicator = row["Indicator"]
            ind = prd.get(Indicator.partial_key(indicator))
            if ind and len(ind) == 1:
                return ind[0]._unit_label if ind[0]._unit_label else ind[0]._unit
            else:
                return ""

        specially_sorted_carriers = sorted(metadata["carriers"], key=len, reverse=True)
        iamc_df["Variable"] = iamc_df.apply(
            functools.partial(_get_iamc_code,
                              prd=state.get("_glb_idx"),
                              iamc_codes=iamc_codes_techs,
                              carriers=specially_sorted_carriers,
                              techs=metadata["techs"]),
            axis=1)
        iamc_df["Unit"] = iamc_df.apply(
            functools.partial(_get_indicator_unit,
                              prd=state.get("_glb_idx")),
            axis=1)

        iamc_df = iamc_df[iamc_df["Variable"] != ""]

        # Rearrange columns
        iamc_df = iamc_df[["Model", "Scenario", "Region", "Variable", "Unit", "Value", "Year", "Processor"]]

        return iamc_df

    # MAIN - Integration, NIS file for the fragment, and NIS submission (Indicators) -----------------------------------
    print(f"Processing fragment {fragment_metadata}")

    # PREPARE / INITIALIZE ------------------
    state = deserialize_state(base_serial_state)
    prd = state.get("_glb_idx")
    dendrogram_musiasem_procs, structural_not_accounted_procs = _get_processors_by_type()

    processors = [
        ["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
         "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong", "Attributes",
         "@SimulationName", "@Region"]]  # TODO Maybe more fields to qualify the fragment
    interfaces = [
        ["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation", "OppositeSubsystemType",
         "GeolocationRef", "GeolocationCode", "InterfaceAttributes", "Value", "Unit", "RelativeTo", "Uncertainty",
         "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
    techs_used_in_regions = set()  # Register if a technology has appeared in a region. Used to generate complete names
    already_added_processors = set()
    already_added_lci = set()
    variables = dict()
    max_lci_interfaces = 100000 if max_lci_interfaces == 0 else max_lci_interfaces
    # If no periods are defined, define a specific year. At least one is needed for the solver to run
    default_time = "2038" if len(fragment_metadata["periods"]) == 0 else "Year"

    # CLONE ------------------
    # Two approaches "clone" or "not clone". "clone" would be to have multi-region models. For now, "not clone"
    clone = False
    cloners = []
    clone_processors = []
    if clone:
        _add_clone(cloners, clone_processors, fragment_metadata["regions"], dendrogram_musiasem_procs)
    else:
        # Modify all parent processors system to region "r"
        if len(fragment_metadata["regions"]) == 1:
            r = next(iter(fragment_metadata["regions"]))
            for p_sim in dendrogram_musiasem_procs.values():
                p_sim.processor_system = r

    # FOR EACH SIMULATION PROCESSOR (of the fragment)
    iamc_codes_techs = create_dictionary()
    for p_sim in fragment_processors:
        # Update time and scenario
        # (they can change from entry to entry, but for MuSIASEM the same functional Processor is used)
        time_ = p_sim.attrs.get("time", p_sim.attrs.get("year", default_time))
        scenario = p_sim.attrs.get("scenario", "")
        region = p_sim.attrs.get("region", "")
        carrier = p_sim.attrs.get("carrier", "")
        tech = p_sim.attrs.get("technology", "")
        subtech = p_sim.attrs.get("subtechnology", g_default_subtech)
        subscenario = p_sim.attrs.get("subscenario", "")
        context = dict(time=time_, scenario=scenario, region=region, carrier=carrier, tech=tech, subtech=subtech)
        if carrier == "":
            print(f"'carrier' is not defined for processor {p_sim.attrs}, ignored")
            continue  # Ignore processors not having carrier defined
        from enbios import subtech_supported
        if subtech_supported and subtech != g_default_subtech:
            # TODO Add subtech support
            print(f"'subtechnology' field is not really supported, processor {p_sim.attrs} ignored")
            continue  # Ignore processors having subtechnology defined

        # Find "LCI" and "Reference Tech" matching Processor(s)
        # (considering "carrier" - "EcoinventCarrierName" if it is defined)
        # matching_lci, matching_not_accounted_reference_tech = _find_lci_and_tech_processors(prd,
        #                                                                                     structural_not_accounted_procs,
        #                                                                                     dendrogram_musiasem_procs,
        #                                                                                     p_sim)
        matching_not_accounted_reference_techs = _find_lci_and_tech_processors(prd,
                                                                               structural_not_accounted_procs,
                                                                               dendrogram_musiasem_procs,
                                                                               p_sim)
        if len(matching_not_accounted_reference_techs) == 0:
            print(f"Could not find a tech processor matching {p_sim.attrs}. Skipping")
            continue

        for matching_not_accounted_reference_tech, matching_lci in matching_not_accounted_reference_techs.items():
            # Parent processor(s) in "MUSIASEM Dendrogram"
            musiasem_matches = _find_parents(prd, structural_not_accounted_procs, matching_not_accounted_reference_tech)
            if musiasem_matches is None or len(musiasem_matches) == 0:
                print(f"{p_sim} has no parents. Cannot account it, skipped.")
                continue

            iamc_codes_techs[tech] = matching_not_accounted_reference_tech.attributes.get(
                "iamccode", matching_not_accounted_reference_tech.full_hierarchy_name.replace(".", "|"))

            # Add PROCESSOR(S)
            name, parent = _add_processor(p_sim, musiasem_matches, processors, already_added_processors, techs_used_in_regions, context)

            # Add INTERFACES (FROM Simulation AND FROM LCI), if we had a processor added (a name)
            if name:  # At least a match? -> add Interfaces, combining
                _add_interfaces(p_sim, name, parent, matching_lci, matching_not_accounted_reference_tech, context)

    print(f"Generating NIS file ...")  # for: {fragment_metadata}")

    # GENERATE NIS file
    nis_file_name = _generate_fragment_nis_file(clone_processors, cloners, processors, interfaces,
                                                variables, fragment_metadata["scenarios"])

    # SUBMIT NIS file: account and calculate indicators, if a NIS file was generated
    outputs = {}
    if nis_file_name:
        print(f'frag_file = "{nis_file_name}"')
        try:
            print(f"Processing NIS file ...")  # for: {fragment_metadata}")
            _ = ["flow_graph_solution_indicators"]
            if generate_interface_results:
                _.append("flow_graph_solution")
            if generate_nis_fragment_file:
                _.append("model")
            _.append("flow_graph_global_indicators")

            issues, new_state, ds = run_nis_for_results(nis_file_name, development_nis_file, state, _)
            if ds:
                outputs["indicators"] = ds[0]
                outputs["iamc_indicators"] = _prepare_iamc_dataframe_fragment(state, ds[0], iamc_codes_techs, fragment_metadata)
                print(f"File processing done, shape of 'Indicators' dataset: {ds[0].shape}, "
                      f"shape of IAMC dataset: {outputs['iamc_indicators'].shape}")
                if generate_interface_results:
                    outputs["interfaces"] = ds[1]
                if generate_nis_fragment_file:
                    outputs["idempotent_nis"] = ds[2] if len(ds) == 3 else ds[1]
                outputs["global_indicators"] = ds[-1]
            else:
                print(f"There were issues processing fragment file: {nis_file_name}")
                print_issues("Solving fragment", nis_file_name, issues, f"Please check the issues resulting from solving fragment '{nis_file_name}'")

        finally:
            if not keep_fragment_file:
                os.remove(nis_file_name)
    return outputs


if __name__ == '__main__':
    base = "https://docs.google.com/spreadsheets/d/1ZXpxYLVO5BXoxLYeJkvfEiOpYxzNwVy01BmB1hnfNQ0/edit?usp=sharing"  # "Copy of Enviro-Sentinel-Base for Script Development"
    base = "https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit?usp=sharing"  # Base full
    base = "https://docs.google.com/spreadsheets/d/1AXCBJZEr8Gw6c9OeCTId3TwUaWZt_ePAOx34HgDJdSI/edit?usp=sharing"  # Base full, LCIA implementation
    # base = "https://docs.google.com/spreadsheets/d/1nYzphq1XYrezquW3yj7UiJ8sNJ8RzJ8gqErg3oVmw2Q/edit?usp=sharing"  # Base with only 1 processor
    frag_file = ""
    if os.path.exists(frag_file):
        # state = deserialize_state(prepare_base_state(base, solve=False))
        state = None
        file, df1, df2 = run_nis_for_results(frag_file, None, state)
        df1.to_csv("~/Downloads/df1.csv", index=False)
        df2.to_csv("~/Downloads/df2.csv", index=False)
    else:
        t = Enviro()
        _ = dict(nis_file_location=base,
                 correspondence_files_path="",
                 simulation_type="sentinel",
                 simulation_files_path="/home/rnebot/Downloads/borrame/calliope-output/datapackage.json",
                 output_directory="/home/rnebot/Downloads/borrame/enviro-output-2/")
        t.set_cfg_file_path(_)
        t.compute_indicators_from_base_and_simulation(n_fragments=True,
                                                      generate_nis_base_file=False,
                                                      generate_nis_fragment_file=False,
                                                      generate_interface_results=True,
                                                      generate_indicators=True,
                                                      max_lci_interfaces=0,
                                                      n_cpus=1)
    print("Done")



from typing import Tuple

import pandas as pd

from enbios.common.helper import isfloat
from enbios.model import SimStructuralProcessorAttributes


def create_register_or_update_processor_and_attributes(prd,
                                                       tech, region, carrier, scenario, time_, subtech, subscenario,
                                                       carrier_idx,
                                                       attrs):
    # tech, region, carrier, scenario, time_, subtech, subscenario -> for col in df.columns
    k = SimStructuralProcessorAttributes.partial_key(tech, region, carrier, scenario, time_,
                                                     subtech, subscenario)
    if carrier_idx >= 0 and not carrier:
        return None
    if tech is None or not region or region == "-":
        # print(f"WARNING: Missing Tech and/or Region: {k}. Skipped")
        return None
    # (tech, region, scenario, time) - (cols)
    o = prd.get(k)
    if len(o) == 0:
        pa = SimStructuralProcessorAttributes(tech, region, carrier, scenario, time_, subtech, subscenario)
        prd.put(k, pa)
    elif len(o) == 1:
        pa = o[0]
    else:
        raise Exception(f"Found {len(o)} occurrences of SimStructuralProcessorAttributes: {k}")
    # Variables from current pd.DataFrame
    _ = pa.attrs
    for key, v in attrs.items():
        if isfloat(v):
            v = float(v)
        if key not in _:
            _[key] = v
        else:
            if isinstance(v, float):
                tmp = _[key]
                _[key] += v  # SUM
                if v != 0 and tmp == _[key]:
                    raise Exception(f"ERROR: did not change. value: {v}, previous: {tmp}")
            else:
                _[key] = v  # Overwrite
    return pa


def find_column_idx_name(columns: pd.Index, possible_values: list) -> Tuple[int, str]:
    if any(columns.isin(possible_values)):
        for p in possible_values:
            try:
                return columns.get_loc(p), p
            except:
                continue
    else:
        return -1, None




conda env list

conda create --name enbios37test --clone enbios37

spolddir =  'C:\Users\1150226\Universitat Autònoma de Barcelona\Sentinel Sostenipra - SENTINEL\SHARED\WP2 -  Socio Environmental\Data\Ecospold\APOS LCI'

pip install enbios==0.7


import copy
import json
import os
from io import BytesIO
from typing import List, Union, BinaryIO
from zipfile import ZipFile

import numpy as np
import pandas as pd
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.writer.excel import save_virtual_workbook

JSON_INDENT = 4
ENSURE_ASCII = False


def _json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    from datetime import datetime

    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    elif isinstance(obj, np.int64):
        return int(obj)
    elif getattr(obj, "Schema"):
        # Use "marshmallow"
        return getattr(obj, "Schema")().dump(obj)
    raise TypeError(f"Type {type(obj)} not serializable")


def generate_json(o):
    return json.dumps(o,
                      default=_json_serial,
                      sort_keys=True,
                      indent=JSON_INDENT,
                      ensure_ascii=ENSURE_ASCII,
                      separators=(',', ': ')
                      ) if o else None


def list_to_dataframe(lst: List) -> pd.DataFrame:
    return pd.DataFrame(data=lst[1:], columns=lst[0])


def get_scenario_name(prefix, s):
    def is_int(si):
        try:
            int(si)
            return True
        except ValueError:
            return False

    return f"{prefix}{s}" if is_int(s) else str(s)


def generate_workbook(cmds, generate_empty=False):
    # Convert list of pd.DataFrames to Excel workbook
    wb = Workbook(write_only=True)
    ws_count = 0
    for name, df in cmds:
        if df.shape[0] == 0:
            if not generate_empty:
                continue

        ws_count += 1
        ws = wb.create_sheet(name)
        widths = [0]*(df.shape[1]+1)  # A maximum of 100 columns
        max_columns = 0
        for r in dataframe_to_rows(df, index=False, header=True):
            if len(r) > max_columns:
                max_columns = len(r)
            for i in range(len(r)):
                width = int(len(str(r[i])) * 1.1)
                if width > widths[i]:
                    widths[i] = width

        for i, column_width in enumerate(widths):
            ws.column_dimensions[get_column_letter(i+1)].width = column_width

        for r in dataframe_to_rows(df, index=False, header=True):
            ws.append(r)

    if ws_count > 0:
        return save_virtual_workbook(wb)
    else:
        return None


def hash_array(f: Union[str, bytes]):
    import hashlib
    m = hashlib.md5()
    if isinstance(f, str):
        m.update(f.encode("utf-8"))
    else:
        m.update(f)
    return m.digest()


def set_zip_timestamp(in_zip: Union[os.PathLike, str, BinaryIO], timestamp=(2020, 1, 1, 0, 0, 0)) -> BinaryIO:
    """
    Modify the timestamp of all files, to stabilize the hash

    :param in_zip: Zip file whose files timestamp will be modified
    :param timestamp: Tuple with the timestamp to set for all the files
    :return A BytesIO with the resulting Zip
    """
    out_zip = BytesIO()
    with ZipFile(in_zip, mode="r") as zin:
        with ZipFile(out_zip, mode="w") as zout:
            for zinfo in zin.infolist():
                data = zin.read(zinfo.filename)
                zinfo_new = copy.copy(zinfo)
                zinfo_new.date_time = timestamp
                zout.writestr(zinfo_new, data)
    return out_zip


def isfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False



import os.path
from typing import List

from bw2io.extractors import Ecospold2DataExtractor


class LCIIndex:
    """ An index of where is each activity, to ease reading it """

    def __init__(self, lci_data_locations: List[str]):
        # TODO Prepare an index of activities, and where each is stored. Maybe also the flows
        self._lci_data_locations = lci_data_locations
        self._activity_locations = {}
        self._activities = {}

    def _read(self):
        self._activities = {}
        self._activity_locations = {}
        for ll in self._lci_data_locations:
            data = Ecospold2DataExtractor.extract(ll, "test", use_mp=False)
            self._activities.update({d["activity"]: d["exchanges"] for d in data})
            self._activity_locations({d["activity"]: ll for d in data})

    def activities(self):
        """
        All indexed activities
        :return:
        """
        return self._activity_locations.keys()

    def activity_location(self, activity):
        return self._activity_locations[activity]

    def read_activity(self, activity):
        """
        Read activity as a dictionary so later it can be inserted into an Interfaces command
        (intensive quantities assumed)

        :param activity:
        :return:
        """
        data = Ecospold2DataExtractor.extract(self.activity_location(activity), "test", use_mp=False)
        return [d["exchanges"] for d in data if d["activity"] == activity][0]
